# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, cast
from urllib.parse import urlparse
from uuid import UUID

import anyio
from structlog import get_logger

from orchestrator.settings import AppSettings, app_settings
from orchestrator.websocket.websocket_manager import WebSocketManager
from orchestrator.workflow import ProcessStatus
from pydantic_forms.types import UUIDstr

logger = get_logger(__name__)


broadcaster_type = urlparse(app_settings.WEBSOCKET_BROADCASTER_URL).scheme


class WS_CHANNELS:
    ALL_PROCESSES = "processes"
    ENGINE_SETTINGS = "engine-settings"
    EVENTS = "events"


async def empty_fn(*args: tuple, **kwargs: dict[str, Any]) -> None:
    return


class WrappedWebSocketManager:
    def __init__(self, wrappee: WebSocketManager | None = None) -> None:
        self.wrapped_websocket_manager = wrappee

    def update(self, wrappee: WebSocketManager) -> None:
        self.wrapped_websocket_manager = wrappee
        if self.wrapped_websocket_manager.enabled:
            logger.info("WebSocketManager object configured, all methods referencing `websocket_manager` should work.")
        else:
            logger.warning("WebSocketManager object not configured, ENABLE_WEBSOCKETS is false.")

    def __getattr__(self, attr: str) -> Any:
        if not isinstance(self.wrapped_websocket_manager, WebSocketManager):
            if "_" in attr:
                logger.warning("No WebSocketManager configured, but attempting to access class methods")
                return None
            raise RuntimeWarning(
                "No WebSocketManager configured at this time. Please pass WebSocketManager configuration to OrchestratorCore base_settings"
            )
        if attr != "enabled" and not self.wrapped_websocket_manager.enabled:
            logger.warning("Websockets are disabled, unable to access class methods")
            return empty_fn

        return getattr(self.wrapped_websocket_manager, attr)


# You need to pass a modified AppSettings class to the OrchestratorCore class to init the WebSocketManager correctly
wrapped_websocket_manager = WrappedWebSocketManager()
websocket_manager = cast(WebSocketManager, wrapped_websocket_manager)


# The Global WebSocketManager is set after calling this function
def init_websocket_manager(settings: AppSettings) -> WebSocketManager:
    wrapped_websocket_manager.update(
        WebSocketManager(settings.ENABLE_WEBSOCKETS, str(settings.WEBSOCKET_BROADCASTER_URL))
    )
    return websocket_manager


def is_process_active(p: dict) -> bool:
    return p["status"] in [ProcessStatus.RUNNING, ProcessStatus.SUSPENDED, ProcessStatus.WAITING]


async def _broadcast_event(name: str, value: Any) -> None:
    event = {"name": name, "value": value}
    await websocket_manager.broadcast_data([WS_CHANNELS.EVENTS], event)


def sync_broadcast_invalidate_cache(cache_object: dict[str, str]) -> None:
    anyio.run(broadcast_invalidate_cache, cache_object)


async def broadcast_invalidate_cache(cache_object: dict[str, str]) -> None:
    await _broadcast_event("invalidateCache", cache_object)


def sync_invalidate_subscription_cache(subscription_id: UUID | UUIDstr, invalidate_all: bool = True) -> None:
    anyio.run(invalidate_subscription_cache, subscription_id, invalidate_all)


async def invalidate_subscription_cache(subscription_id: UUID | UUIDstr, invalidate_all: bool = True) -> None:
    if invalidate_all:
        await broadcast_invalidate_cache({"type": "subscriptions"})
    await broadcast_invalidate_cache({"type": "subscriptions", "id": "LIST"})
    await broadcast_invalidate_cache({"type": "subscriptions", "id": str(subscription_id)})


def broadcast_invalidate_status_counts() -> None:
    """Broadcast message to invalidate the status counts of the connected websocket clients."""
    if not websocket_manager.enabled:
        logger.debug("WebSocketManager is not enabled. Skip broadcasting through websocket.")
        return

    sync_broadcast_invalidate_cache({"type": "processStatusCounts"})


def broadcast_process_update_to_websocket(
    process_id: UUID,
) -> None:
    """Broadcast data of the current process to connected websocket clients."""
    if not websocket_manager.enabled:
        logger.debug(
            "WebSocketManager is not enabled. Skip broadcasting through websocket.", process_id=str(process_id)
        )
        return

    sync_broadcast_invalidate_cache({"type": "processes", "id": "LIST"})
    sync_broadcast_invalidate_cache({"type": "processes", "id": str(process_id)})


async def broadcast_process_update_to_websocket_async(
    process_id: UUID,
) -> None:
    if not websocket_manager.enabled:
        logger.debug(
            "WebSocketManager is not enabled. Skip broadcasting through websocket.", process_id=str(process_id)
        )
        return

    await broadcast_invalidate_cache({"type": "processes", "id": "LIST"})
    await broadcast_invalidate_cache({"type": "processes", "id": str(process_id)})


__all__ = [
    "websocket_manager",
    "init_websocket_manager",
    "broadcast_process_update_to_websocket",
    "broadcast_process_update_to_websocket_async",
    "WS_CHANNELS",
    "broadcast_invalidate_status_counts",
]



from orchestrator.workflows import LazyWorkflowInstance

LazyWorkflowInstance("workflows.example2.create_example2", "create_example2")
LazyWorkflowInstance("workflows.example2.modify_example2", "modify_example2")
LazyWorkflowInstance("workflows.example2.terminate_example2", "terminate_example2")
LazyWorkflowInstance("workflows.example2.validate_example2", "validate_example2")
LazyWorkflowInstance("workflows.example1.create_example1", "create_example1")
LazyWorkflowInstance("workflows.example1.modify_example1", "modify_example1")
LazyWorkflowInstance("workflows.example1.terminate_example1", "terminate_example1")
LazyWorkflowInstance("workflows.example1.validate_example1", "validate_example1")
LazyWorkflowInstance("workflows.example4.create_example4", "create_example4")
LazyWorkflowInstance("workflows.example4.modify_example4", "modify_example4")
LazyWorkflowInstance("workflows.example4.terminate_example4", "terminate_example4")
LazyWorkflowInstance("workflows.example4.validate_example4", "validate_example4")



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import structlog

from orchestrator.api.helpers import add_subscription_search_query_filter
from orchestrator.db import ProductTable, SubscriptionTable
from orchestrator.db.filters import create_memoized_field_list, generic_filter_from_clauses
from orchestrator.db.filters.search_filters import default_inferred_column_clauses, inferred_filter
from orchestrator.db.filters.search_filters.inferred_filter import filter_exact

logger = structlog.get_logger(__name__)


SUBSCRIPTION_TABLE_COLUMN_CLAUSES = default_inferred_column_clauses(SubscriptionTable) | {
    "product": inferred_filter(ProductTable.name),
    "tag": filter_exact(ProductTable.tag),
    "type": filter_exact(ProductTable.product_type),
}

subscription_filter_fields = create_memoized_field_list(SUBSCRIPTION_TABLE_COLUMN_CLAUSES)
filter_subscriptions = generic_filter_from_clauses(SUBSCRIPTION_TABLE_COLUMN_CLAUSES)

filter_by_query_string = add_subscription_search_query_filter



import random
import string
from typing import Any

from orchestrator.db import (
    ProductTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.utils.datetime import nowtz
from pydantic_forms.types import SubscriptionMapping


def randomstr(length=3):
    charset = string.ascii_lowercase + string.digits
    return "".join(random.choice(charset) for _ in range(0, length))  # noqa: S311


def create_subscription_for_mapping(
    product: ProductTable, mapping: SubscriptionMapping, values: dict[str, Any], **kwargs: Any
) -> SubscriptionTable:
    """Create a subscription in the test coredb for the given subscription_mapping and values.

    This function handles optional resource types starting with a ? in the mapping not supplied in the values array.

    Args:
        product: the ProductTable to create a sub for
        mapping: the subscription_mapping belonging to that product
        values: a dictionary of keys from the sub_map and their corresponding test values
        kwargs: The rest of the arguments

    Returns: The conforming subscription.
    """

    def build_instance(name, value_mapping):
        block = product.find_block_by_name(name)

        def build_value(rt, value):
            resource_type = block.find_resource_type_by_name(rt)
            return SubscriptionInstanceValueTable(resource_type_id=resource_type.resource_type_id, value=value)

        return SubscriptionInstanceTable(
            product_block_id=block.product_block_id,
            values=[
                build_value(resource_type, values[value_key]) for (resource_type, value_key) in value_mapping.items()
            ],
        )

    # recreate the mapping: leave out the ?keys if no value supplied for them
    mapping = {
        name: [
            {
                **{k: value_map[k] for k in value_map if not value_map[k].startswith("?")},
                **{
                    k: value_map[k][1:]
                    for k in value_map
                    if value_map[k].startswith("?") and value_map[k][1:] in values
                },
            }
            for value_map in mapping[name]
        ]
        for name in mapping
    }

    instances = [
        build_instance(name, value_mapping)
        for (name, value_mappings) in mapping.items()
        for value_mapping in value_mappings
    ]

    return create_subscription(instances=instances, product=product, **kwargs)


def create_subscription(**kwargs):
    attrs = {
        "description": "A subscription.",
        "customer_id": "85938c4c-0a11-e511-80d0-005056956c1a",
        "start_date": nowtz(),
        "status": "active",
        "insync": True,
        **kwargs,
    }
    o = SubscriptionTable(**attrs)
    db.session.add(o)
    db.session.commit()
    return o



# Copyright 2019-2025 GÉANT, SURF, ESnet
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from orchestrator.db import db
from orchestrator.db.models import InputStateTable, ProcessTable


@pytest.fixture
def create_input_state():
    """Fixture to create an input state."""

    def _create_input_state(process_id, input_state, input_type, **kwargs):
        input_state_entry = InputStateTable(
            process_id=process_id, input_state=input_state, input_type=input_type, **kwargs
        )
        db.session.add(input_state_entry)
        db.session.commit()
        return input_state_entry

    return _create_input_state


def test_cascade_delete(mocked_processes, test_client, create_input_state):
    # Create a process and related input state
    processes = test_client.get("/api/processes").json()
    # Create an input state for the first process
    create_input_state(
        process_id=processes[0]["process_id"],
        input_state={"key": "value"},
        input_type="initial_state",
    )

    # Verify both records exist
    assert db.session.query(ProcessTable).count() == 9
    assert db.session.query(InputStateTable).count() == 1

    # Delete one of the process so that the input state is deleted as well
    process = db.session.query(ProcessTable).filter_by(process_id=processes[0]["process_id"]).one()
    db.session.delete(process)
    db.session.commit()

    # Verify cascade delete
    assert db.session.query(ProcessTable).count() == 8
    assert db.session.query(InputStateTable).count() == 0



from uuid import uuid4

import pytest

from orchestrator.db import db
from orchestrator.db.models import FixedInputTable, ProductTable
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import ProductModel, SubscriptionModel
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_product_model_list_nested_product_type_one(
    test_product_block_list_nested_db_in_use_by_block, test_product_type_list_nested
):
    product = ProductTable(
        name="TestProductListNestedTypeOne",
        description="Test ProductTable ProductTypeOne",
        product_type="ProductTypeOne",
        tag="TEST",
        status="active",
    )

    fixed_input = FixedInputTable(name="test_fixed_input", value="1")

    product_block = test_product_block_list_nested_db_in_use_by_block
    product.fixed_inputs = [fixed_input]
    product.product_blocks = [product_block]

    db.session.add(product)
    db.session.commit()

    _, _, ProductTypeListNestedForTest = test_product_type_list_nested
    SUBSCRIPTION_MODEL_REGISTRY["TestProductListNestedTypeOne"] = ProductTypeListNestedForTest
    yield ProductModel(
        product_id=product.product_id,
        name="TestProductListNestedTypeOne",
        description=product.description,
        product_type=product.product_type,
        tag=product.tag,
        status=ProductLifecycle.ACTIVE,
    )
    del SUBSCRIPTION_MODEL_REGISTRY["TestProductListNestedTypeOne"]


@pytest.fixture
def test_product_model_list_nested_product_type_two(
    test_product_block_list_nested_db_in_use_by_block, test_product_type_list_nested
):
    product = ProductTable(
        name="TestProductListNestedType2",
        description="Test ProductTable ProductTypeTwo",
        product_type="ProductTypeTwo",
        tag="TEST",
        status="active",
    )

    fixed_input = FixedInputTable(name="test_fixed_input", value="True")

    product_block = test_product_block_list_nested_db_in_use_by_block
    product.fixed_inputs = [fixed_input]
    product.product_blocks = [product_block]

    db.session.add(product)
    db.session.commit()

    _, _, ProductTypeListNestedForTest = test_product_type_list_nested
    SUBSCRIPTION_MODEL_REGISTRY["TestProductListNestedType2"] = ProductTypeListNestedForTest
    yield ProductModel(
        product_id=product.product_id,
        name="TestProductListNestedType2",
        description=product.description,
        product_type=product.product_type,
        tag=product.tag,
        status=ProductLifecycle.ACTIVE,
    )
    del SUBSCRIPTION_MODEL_REGISTRY["TestProductListNestedType2"]


@pytest.fixture
def factory_subscription_with_nestings_in_use_by(
    test_product_block_list_nested,
    test_product_model_list_nested,
    test_product_type_list_nested,
    test_product_block_list_nested_db_in_use_by_block,
    test_product_model_list_nested_product_type_one,
    test_product_model_list_nested_product_type_two,
):
    """Fixture that creates subscriptions with multiple nestings.

    relations for in use by:
    - subscription_10
        - subscription_20
            - subscription_30
                - subscription_40
                - subscription_41 - terminated
            - subscription_31 - terminated
                - subscription_42
            - subscription_32 - type: ProductTypeOne
                - subscription_40
                - subscription_43

        - subscription_21 - terminated
            - subscription_33
                - subscription_44

        - subscription_22 - type: ProductTypeOne
            - subscription_34 - type: ProductTypeTwo
                - subscription_45

    Returns: dict with subscription ids
    """

    ProductTypeListNestedForTestInactive, _, _ = test_product_type_list_nested
    ProductBlockListNestedForTestInactive, _, _ = test_product_block_list_nested

    customer_id = str(uuid4())

    def create_subscription(*, int_value, sub_blocks=(), different_product_type=None, is_terminated=False):
        product_id = test_product_model_list_nested.product_id
        if different_product_type:
            product_id = different_product_type.product_id

        subscription = ProductTypeListNestedForTestInactive.from_product_id(
            product_id=product_id, customer_id=customer_id, insync=True
        )
        subscription.block = ProductBlockListNestedForTestInactive.new(
            subscription_id=subscription.subscription_id, int_field=int_value, sub_block_list=list(sub_blocks)
        )

        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        if is_terminated:
            subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.TERMINATED)

        subscription.save()
        db.session.commit()
        return subscription

    subscription_10 = create_subscription(int_value=10)

    subscription_20 = create_subscription(int_value=20, sub_blocks=[subscription_10.block])
    subscription_21 = create_subscription(int_value=21, sub_blocks=[subscription_10.block], is_terminated=True)
    subscription_22 = create_subscription(
        int_value=22,
        sub_blocks=[subscription_10.block],
        different_product_type=test_product_model_list_nested_product_type_one,
    )

    subscription_30 = create_subscription(int_value=30, sub_blocks=[subscription_20.block])
    subscription_31 = create_subscription(int_value=31, sub_blocks=[subscription_20.block], is_terminated=True)
    subscription_32 = create_subscription(
        int_value=32,
        sub_blocks=[subscription_20.block],
        different_product_type=test_product_model_list_nested_product_type_one,
    )
    subscription_33 = create_subscription(int_value=33, sub_blocks=[subscription_21.block])
    subscription_34 = create_subscription(
        int_value=34,
        sub_blocks=[subscription_22.block],
        different_product_type=test_product_model_list_nested_product_type_two,
    )

    subscription_40 = create_subscription(int_value=40, sub_blocks=[subscription_30.block, subscription_32.block])
    subscription_41 = create_subscription(int_value=41, sub_blocks=[subscription_30.block], is_terminated=True)
    subscription_42 = create_subscription(int_value=42, sub_blocks=[subscription_31.block])
    subscription_43 = create_subscription(int_value=43, sub_blocks=[subscription_32.block])
    subscription_44 = create_subscription(int_value=44, sub_blocks=[subscription_33.block])
    subscription_45 = create_subscription(int_value=45, sub_blocks=[subscription_34.block])

    return {
        "subscription_10": subscription_10.subscription_id,
        "subscription_20": subscription_20.subscription_id,
        "subscription_21": subscription_21.subscription_id,
        "subscription_22": subscription_22.subscription_id,
        "subscription_30": subscription_30.subscription_id,
        "subscription_31": subscription_31.subscription_id,
        "subscription_32": subscription_32.subscription_id,
        "subscription_33": subscription_33.subscription_id,
        "subscription_34": subscription_34.subscription_id,
        "subscription_40": subscription_40.subscription_id,
        "subscription_41": subscription_41.subscription_id,
        "subscription_42": subscription_42.subscription_id,
        "subscription_43": subscription_43.subscription_id,
        "subscription_44": subscription_44.subscription_id,
        "subscription_45": subscription_45.subscription_id,
    }


@pytest.fixture
def factory_subscription_with_nestings_depends_on(
    test_product_block_list_nested,
    test_product_model_list_nested,
    test_product_type_list_nested,
    test_product_block_list_nested_db_in_use_by_block,
    test_product_model_list_nested_product_type_one,
    test_product_model_list_nested_product_type_two,
):
    """Fixture that creates subscriptions with multiple nestings.

    relations for depends on (default product type `Test`):
    - subscription_40
        - subscription_30
            - subscription_20
                - subscription_10
                - subscription_11 - terminated
            - subscription_21 - terminated
                - subscription_12
            - subscription_22 - type: ProductTypeOne
                - subscription_10
                - subscription_13

        - subscription_31 - terminated
            - subscription_23
                - subscription_14

        - subscription_32 - type: ProductTypeOne
            - subscription_24 - type: ProductTypeTwo
                - subscription_15

    Returns: dict with subscription ids
    """
    ProductTypeListNestedForTestInactive, _, _ = test_product_type_list_nested
    ProductBlockListNestedForTestInactive, _, _ = test_product_block_list_nested

    customer_id = str(uuid4())

    def create_subscription(*, int_value, sub_blocks=(), different_product_type=None, is_terminated=False):
        product_id = test_product_model_list_nested.product_id
        if different_product_type:
            product_id = different_product_type.product_id

        subscription = ProductTypeListNestedForTestInactive.from_product_id(
            product_id=product_id, customer_id=customer_id, insync=True
        )
        subscription.block = ProductBlockListNestedForTestInactive.new(
            subscription_id=subscription.subscription_id, int_field=int_value, sub_block_list=list(sub_blocks)
        )

        subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.ACTIVE)
        if is_terminated:
            subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.TERMINATED)

        subscription.save()
        db.session.commit()
        return subscription

    subscription_10 = create_subscription(int_value=10)
    subscription_11 = create_subscription(int_value=11, is_terminated=True)
    subscription_12 = create_subscription(int_value=12)
    subscription_13 = create_subscription(int_value=13)
    subscription_14 = create_subscription(int_value=14)
    subscription_15 = create_subscription(int_value=15)

    subscription_20 = create_subscription(int_value=20, sub_blocks=[subscription_10.block, subscription_11.block])
    subscription_21 = create_subscription(int_value=21, sub_blocks=[subscription_12.block], is_terminated=True)
    subscription_22 = create_subscription(
        int_value=22,
        sub_blocks=[subscription_10.block, subscription_13.block],
        different_product_type=test_product_model_list_nested_product_type_one,
    )
    subscription_23 = create_subscription(int_value=23, sub_blocks=[subscription_14.block])
    subscription_24 = create_subscription(
        int_value=24,
        sub_blocks=[subscription_15.block],
        different_product_type=test_product_model_list_nested_product_type_two,
    )

    subscription_30 = create_subscription(
        int_value=30, sub_blocks=[subscription_20.block, subscription_21.block, subscription_22.block]
    )
    subscription_31 = create_subscription(int_value=31, sub_blocks=[subscription_23.block], is_terminated=True)
    subscription_32 = create_subscription(
        int_value=32,
        sub_blocks=[subscription_24.block],
        different_product_type=test_product_model_list_nested_product_type_one,
    )

    # create subscription 40 that use subscription 30, 31 and 32
    subscription_40 = create_subscription(
        int_value=40, sub_blocks=[subscription_30.block, subscription_31.block, subscription_32.block]
    )

    return {
        "subscription_10": subscription_10.subscription_id,
        "subscription_11": subscription_11.subscription_id,
        "subscription_12": subscription_12.subscription_id,
        "subscription_13": subscription_13.subscription_id,
        "subscription_14": subscription_14.subscription_id,
        "subscription_15": subscription_15.subscription_id,
        "subscription_20": subscription_20.subscription_id,
        "subscription_21": subscription_21.subscription_id,
        "subscription_22": subscription_22.subscription_id,
        "subscription_23": subscription_23.subscription_id,
        "subscription_24": subscription_24.subscription_id,
        "subscription_30": subscription_30.subscription_id,
        "subscription_31": subscription_31.subscription_id,
        "subscription_32": subscription_32.subscription_id,
        "subscription_40": subscription_40.subscription_id,
    }



from unittest import mock

from orchestrator.graphql.utils.get_selected_paths import get_selected_paths
from test.unit_tests.graphql.utils.fixtures import LARGE_QUERY_SELECTED_FIELDS, SMALL_QUERY_SELECTED_FIELDS


def test_get_selected_paths_large():
    mock_info = mock.Mock()
    mock_info.selected_fields = LARGE_QUERY_SELECTED_FIELDS
    paths = get_selected_paths(mock_info)

    assert sorted(paths) == sorted(
        [
            "page.name",
            "page.target",
            "page.products.tag",
            "page.products.product_blocks.name",
            "page.products.product_blocks.resource_types.description",
            "page.products.product_blocks.resource_types.product_blocks.name",
            "page.products.product_blocks.in_use_by.name",
            "page.created_at",
        ]
    )


def test_get_selected_paths_empty():
    # given
    mock_info = mock.Mock()
    mock_info.selected_fields = []

    # when
    paths = get_selected_paths(mock_info)

    # then
    assert paths == []


def test_get_selected_paths_one():
    # given
    mock_info = mock.Mock()
    mock_info.selected_fields = SMALL_QUERY_SELECTED_FIELDS

    # when
    paths = get_selected_paths(mock_info)

    # then
    assert paths == [
        "page.created_at",
    ]



import json

from orchestrator.db import SubscriptionCustomerDescriptionTable, db
from test.unit_tests.config import GRAPHQL_ENDPOINT, GRAPHQL_HEADERS
from test.unit_tests.graphql.mutations.helpers import mutation_authorization


def get_customer_description_upsert_mutation(
    customer_id: str,
    subscription_id: str,
    description: str,
    version=None,
) -> bytes:
    query = """
mutation CustomerDescriptionUpsertMutation ($customerId: String!, $subscriptionId: UUID!, $description: String!, $version: Int) {
    upsertCustomerDescription(customerId: $customerId, subscriptionId: $subscriptionId, description: $description, version: $version) {
        ... on CustomerDescription {
            customerId
            subscriptionId
            description
            version
        }

        ... on NotFoundError {
            message
        }
        ... on MutationError {
            message
        }
    }
}
    """
    return json.dumps(
        {
            "operationName": "CustomerDescriptionUpsertMutation",
            "query": query,
            "variables": {
                "customerId": customer_id,
                "subscriptionId": subscription_id,
                "description": description,
                "version": version,
            },
        }
    ).encode("utf-8")


def get_customer_description_remove_mutation(
    customer_id: str,
    subscription_id: str,
) -> bytes:
    query = """
mutation CustomerDescriptionRemoveMutation ($customerId: String!, $subscriptionId: UUID!) {
    removeCustomerDescription(customerId: $customerId, subscriptionId: $subscriptionId) {
        ... on CustomerDescription {
            customerId
            subscriptionId
            description
            version
        }

        ... on NotFoundError {
            message
        }
    }
}
    """
    return json.dumps(
        {
            "operationName": "CustomerDescriptionRemoveMutation",
            "query": query,
            "variables": {
                "customerId": customer_id,
                "subscriptionId": subscription_id,
            },
        }
    ).encode("utf-8")


def test_customer_description_create(httpx_mock, test_client, product_sub_list_union_subscription_1):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"
    description = "create description"

    # when

    query = get_customer_description_upsert_mutation(customer_id, subscription_id, description)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "upsertCustomerDescription": {
                "customerId": "0c43b714-0a11-e511-80d0-005056956c1a",
                "subscriptionId": subscription_id,
                "description": "create description",
                "version": 1,
            }
        }
    }


def test_customer_description_create_not_found_error(httpx_mock, test_client):
    # given

    subscription_id = "50b44ca0-25f6-40d4-b2af-f0418bf2e81a"
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"
    description = "create description"

    # when

    query = get_customer_description_upsert_mutation(customer_id, subscription_id, description)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "upsertCustomerDescription": {
                "message": "Subscription not found",
            }
        }
    }


def test_customer_description_update(httpx_mock, test_client, product_sub_list_union_subscription_1):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"
    description = "update description"

    customer_description = SubscriptionCustomerDescriptionTable(
        customer_id=customer_id,
        subscription_id=subscription_id,
        description="create description",
    )
    db.session.add(customer_description)
    db.session.commit()

    # when
    query = get_customer_description_upsert_mutation(customer_id, subscription_id, description)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "upsertCustomerDescription": {
                "customerId": "0c43b714-0a11-e511-80d0-005056956c1a",
                "subscriptionId": subscription_id,
                "description": "update description",
                "version": 2,
            }
        }
    }


def test_customer_description_update_with_version(httpx_mock, test_client, product_sub_list_union_subscription_1):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"
    description = "update description"
    version = 1

    customer_description = SubscriptionCustomerDescriptionTable(
        customer_id=customer_id,
        subscription_id=subscription_id,
        description="create description",
        version=version,
    )
    db.session.add(customer_description)
    db.session.commit()

    # when
    query = get_customer_description_upsert_mutation(customer_id, subscription_id, description, version)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "upsertCustomerDescription": {
                "customerId": "0c43b714-0a11-e511-80d0-005056956c1a",
                "subscriptionId": subscription_id,
                "description": "update description",
                "version": version + 1,
            }
        }
    }


def test_customer_description_update_with_incorrect_version(
    httpx_mock, test_client, product_sub_list_union_subscription_1
):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"
    description = "update description"
    version = 0

    customer_description = SubscriptionCustomerDescriptionTable(
        customer_id=customer_id,
        subscription_id=subscription_id,
        description="create description",
        version=1,
    )
    db.session.add(customer_description)
    db.session.commit()

    # when
    query = get_customer_description_upsert_mutation(customer_id, subscription_id, description, version)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "upsertCustomerDescription": {
                "message": "Stale data: given version (0) does not match the current version (1)",
            },
        },
    }


def test_customer_description_delete(httpx_mock, test_client, product_sub_list_union_subscription_1):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"

    customer_description = SubscriptionCustomerDescriptionTable(
        customer_id=customer_id,
        subscription_id=subscription_id,
        description="create description",
    )
    db.session.add(customer_description)
    db.session.commit()

    # when

    query = get_customer_description_remove_mutation(customer_id, subscription_id)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "removeCustomerDescription": {
                "customerId": "0c43b714-0a11-e511-80d0-005056956c1a",
                "subscriptionId": subscription_id,
                "description": "create description",
                "version": 1,
            }
        }
    }


def test_customer_description_delete_not_found_error(httpx_mock, test_client, product_sub_list_union_subscription_1):
    # given

    subscription_id = str(product_sub_list_union_subscription_1)
    customer_id = "0c43b714-0a11-e511-80d0-005056956c1a"

    # when

    query = get_customer_description_remove_mutation(customer_id, subscription_id)

    with mutation_authorization():
        response = test_client.post(GRAPHQL_ENDPOINT, content=query, headers=GRAPHQL_HEADERS)

    # then

    assert response.json() == {
        "data": {
            "removeCustomerDescription": {
                "message": "Customer description not found",
            }
        }
    }



from uuid import uuid4

import pytest
from pydantic import TypeAdapter, ValidationError

from orchestrator.forms.validators import customer_contact_list
from pydantic_forms.core import FormPage


def test_customer_contact_list():
    ContactPersonList = customer_contact_list()

    class Form(FormPage):
        customer_contacts: ContactPersonList

    validated_data = Form(
        customer_contacts=[{"name": "test1", "email": "a@b.nl", "phone": ""}, {"name": "test2", "email": "a@b.nl"}]
    ).model_dump()

    expected = {
        "customer_contacts": [
            {"email": "a@b.nl", "name": "test1", "phone": ""},
            {"email": "a@b.nl", "name": "test2", "phone": ""},
        ]
    }
    assert validated_data == expected


def test_customer_contact_list_schema():
    customer_id = uuid4()
    customer_id_str = str(customer_id)

    CustomerContactList = customer_contact_list(customer_id, "key", min_items=1)

    class Form(FormPage):
        customer_contacts: customer_contact_list()
        customer_contacts_customer: CustomerContactList
        customer_contacts_customer2: customer_contact_list(customer_id, "foo")  # noqa: F821

    expected = {
        "$defs": {
            "ContactPerson": {
                "properties": {
                    "name": {"format": "contactPersonName", "title": "Name", "type": "string"},
                    "email": {"format": "email", "title": "Email", "type": "string"},
                    "phone": {"default": "", "title": "Phone", "type": "string"},
                },
                "required": ["name", "email"],
                "title": "ContactPerson",
                "type": "object",
            }
        },
        "additionalProperties": False,
        "properties": {
            "customer_contacts": {
                "customerKey": "customer_id",
                "items": {"$ref": "#/$defs/ContactPerson"},
                "title": "Customer Contacts",
                "type": "array",
            },
            "customer_contacts_customer": {
                "customerId": customer_id_str,
                "customerKey": "key",
                "items": {"$ref": "#/$defs/ContactPerson"},
                "minItems": 1,
                "title": "Customer Contacts Customer",
                "type": "array",
            },
            "customer_contacts_customer2": {
                "customerId": customer_id_str,
                "customerKey": "foo",
                "items": {"$ref": "#/$defs/ContactPerson"},
                "title": "Customer Contacts Customer2",
                "type": "array",
            },
        },
        "required": ["customer_contacts", "customer_contacts_customer", "customer_contacts_customer2"],
        "title": "unknown",
        "type": "object",
    }

    assert Form.model_json_schema() == expected


@pytest.fixture(name="Form")
def form_with_customer_contact_list():
    customer_id = uuid4()

    ReqContactList = customer_contact_list(min_items=1)
    CustomerContactList = customer_contact_list(customer_id=customer_id, customer_key="key")

    class Form(FormPage):
        customer_contacts: ReqContactList
        customer_contacts_customer: CustomerContactList = []

    return Form


def test_customer_contact_list_nok_invalid_email(Form):
    with pytest.raises(ValidationError) as error_info:
        Form(customer_contacts=[{"name": "test1", "email": "a@b"}, {"email": "a@b.nl"}])

    errors = error_info.value.errors(include_url=False, include_context=False)
    expected = [
        {
            "input": "a@b",
            "loc": ("customer_contacts", 0, "email"),
            "msg": "value is not a valid email address: The part after the @-sign is not valid. It should have a period.",
            "type": "value_error",
        },
        {
            "input": {"email": "a@b.nl"},
            "loc": ("customer_contacts", 1, "name"),
            "msg": "Field required",
            "type": "missing",
        },
    ]
    assert errors == expected


def test_customer_contact_list_nok_empty(Form):
    with pytest.raises(ValidationError) as error_info:
        Form(customer_contacts=[])

    errors = error_info.value.errors(include_url=False)
    expected = [
        {
            "input": [],
            "loc": ("customer_contacts",),
            "msg": "List should have at least 1 item after validation, not 0",
            "type": "too_short",
            "ctx": {"actual_length": 0, "field_type": "List", "min_length": 1},
        }
    ]
    assert errors == expected


@pytest.mark.parametrize("value", [[], [{}]])
def test_customer_contact_list_empty_values(value):
    assert TypeAdapter(customer_contact_list()).validate_python(value) == []



# Copyright 2019-2025 SURF, GÉANT, ESnet.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements process related API endpoints."""

import struct
import zlib
from http import HTTPStatus
from typing import Any
from uuid import UUID

import structlog
from fastapi import Request
from fastapi.param_functions import Body, Depends, Header
from fastapi.routing import APIRouter
from fastapi.websockets import WebSocket
from fastapi_etag.dependency import CacheHit
from more_itertools import chunked
from sentry_sdk.tracing import trace
from sqlalchemy import CompoundSelect, Select, select
from sqlalchemy.orm import defer, joinedload
from sqlalchemy.sql.functions import count
from starlette.responses import Response

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.api.error_handling import raise_status
from orchestrator.api.helpers import add_response_range
from orchestrator.db import ProcessSubscriptionTable, ProcessTable, SubscriptionTable, db
from orchestrator.db.filters import Filter
from orchestrator.db.filters.process import filter_processes
from orchestrator.db.sorting import Sort, SortOrder
from orchestrator.db.sorting.process import sort_processes
from orchestrator.schemas import ProcessIdSchema, ProcessResumeAllSchema, ProcessSchema, ProcessStatusCounts, Reporter
from orchestrator.security import authenticate
from orchestrator.services.process_broadcast_thread import api_broadcast_process_data
from orchestrator.services.processes import (
    SYSTEM_USER,
    _async_resume_processes,
    _get_process,
    abort_process,
    continue_awaiting_process,
    load_process,
    resume_process,
    start_process,
    update_awaiting_process_progress,
)
from orchestrator.services.settings import get_engine_settings
from orchestrator.settings import app_settings
from orchestrator.utils.enrich_process import enrich_process
from orchestrator.websocket import (
    WS_CHANNELS,
    broadcast_invalidate_status_counts,
    broadcast_process_update_to_websocket,
    websocket_manager,
)
from orchestrator.workflow import ProcessStatus
from pydantic_forms.types import JSON, State

router = APIRouter()

logger = structlog.get_logger(__name__)


def check_global_lock() -> None:
    """Check the global lock of the engine.

    Returns:
        None or raises an exception

    """
    engine_settings = get_engine_settings()
    if engine_settings.global_lock:
        logger.info("Unable to interact with processes at this time. Engine StatusEnum is locked")
        raise_status(
            HTTPStatus.SERVICE_UNAVAILABLE, detail="Engine is locked cannot accept changes on processes at this time"
        )


def resolve_user_name(
    *,
    reporter: Reporter | None,
    resolved_user: OIDCUserModel | None,
) -> str:
    if reporter:
        return reporter

    if resolved_user:
        return resolved_user.name if resolved_user.name else resolved_user.user_name

    return SYSTEM_USER


def user_name(
    reporter: Reporter | None = None,
    user: OIDCUserModel | None = Depends(authenticate),
) -> str:
    return resolve_user_name(reporter=reporter, resolved_user=user)


@router.delete("/{process_id}", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def delete(process_id: UUID) -> None:
    stmt = select(ProcessTable).filter_by(process_id=process_id)
    process = db.session.execute(stmt).scalar_one_or_none()

    if not process:
        raise_status(HTTPStatus.NOT_FOUND)

    db.session.delete(db.session.get(ProcessTable, process_id))
    db.session.commit()

    broadcast_invalidate_status_counts()
    broadcast_process_update_to_websocket(process.process_id)


@router.post(
    "/{workflow_key}",
    response_model=ProcessIdSchema,
    status_code=HTTPStatus.CREATED,
    dependencies=[Depends(check_global_lock, use_cache=False)],
)
def new_process(
    workflow_key: str,
    request: Request,
    json_data: list[dict[str, Any]] | None = Body(...),
    user: str = Depends(user_name),
    user_model: OIDCUserModel | None = Depends(authenticate),
) -> dict[str, UUID]:
    broadcast_func = api_broadcast_process_data(request)
    process_id = start_process(
        workflow_key, user_inputs=json_data, user_model=user_model, user=user, broadcast_func=broadcast_func
    )

    return {"id": process_id}


@router.put(
    "/{process_id}/resume",
    response_model=None,
    status_code=HTTPStatus.NO_CONTENT,
    dependencies=[Depends(check_global_lock, use_cache=False)],
)
def resume_process_endpoint(
    process_id: UUID, request: Request, json_data: JSON = Body(...), user: str = Depends(user_name)
) -> None:
    process = _get_process(process_id)

    if process.last_status == ProcessStatus.COMPLETED:
        raise_status(HTTPStatus.CONFLICT, "Resuming a completed workflow is not possible")

    if process.last_status == ProcessStatus.RUNNING:
        raise_status(HTTPStatus.CONFLICT, "Resuming a running workflow is not possible")

    if process.last_status == ProcessStatus.RESUMED:
        raise_status(HTTPStatus.CONFLICT, "Resuming a resumed workflow is not possible")

    broadcast_invalidate_status_counts()
    broadcast_func = api_broadcast_process_data(request)

    resume_process(process, user=user, user_inputs=json_data, broadcast_func=broadcast_func)


@router.post(
    "/{process_id}/callback/{token}",
    response_model=None,
    status_code=HTTPStatus.OK,
    dependencies=[Depends(check_global_lock, use_cache=False)],
)
def continue_awaiting_process_endpoint(
    process_id: UUID,
    token: str,
    request: Request,
    json_data: State = Body(...),
) -> None:
    check_global_lock()

    process = _get_process(process_id)

    if process.last_status != ProcessStatus.AWAITING_CALLBACK:
        raise_status(HTTPStatus.CONFLICT, "This process is not in an awaiting state.")

    try:
        broadcast_func = api_broadcast_process_data(request)
        continue_awaiting_process(process, token=token, input_data=json_data, broadcast_func=broadcast_func)
    except AssertionError as e:
        raise_status(HTTPStatus.NOT_FOUND, str(e))


@router.post(
    "/{process_id}/callback/{token}/progress",
    response_model=None,
    status_code=HTTPStatus.OK,
    dependencies=[Depends(check_global_lock, use_cache=False)],
)
def update_progress_on_awaiting_process_endpoint(
    process_id: UUID,
    token: str,
    data: str | State = Body(...),
) -> None:
    process = _get_process(process_id)

    if process.last_status != ProcessStatus.AWAITING_CALLBACK:
        raise_status(HTTPStatus.CONFLICT, "This process is not in an awaiting state.")

    try:
        update_awaiting_process_progress(process, token=token, data=data)
    except AssertionError as exc:
        raise_status(HTTPStatus.NOT_FOUND, str(exc))


@router.put(
    "/resume-all", response_model=ProcessResumeAllSchema, dependencies=[Depends(check_global_lock, use_cache=False)]
)
async def resume_all_processess_endpoint(request: Request, user: str = Depends(user_name)) -> dict[str, int]:
    """Retry all task processes in status Failed, Waiting, API Unavailable or Inconsistent Data.

    The retry is started in the background, returning status 200 and number of processes in message.
    When it is already running, refuse and return status 409 instead.
    """

    # Retrieve processes eligible for resuming
    stmt = (
        select(ProcessTable)
        .filter(
            ProcessTable.last_status.in_(
                [
                    ProcessStatus.FAILED,
                    ProcessStatus.WAITING,
                    ProcessStatus.API_UNAVAILABLE,
                    ProcessStatus.INCONSISTENT_DATA,
                ]
            )
        )
        .filter(ProcessTable.is_task.is_(True))
    )
    processes_to_resume = db.session.scalars(stmt).all()

    broadcast_func = api_broadcast_process_data(request)
    if not await _async_resume_processes(processes_to_resume, user, broadcast_func=broadcast_func):
        raise_status(HTTPStatus.CONFLICT, "Another request to resume all processes is in progress")

    logger.info("Resuming all processes", count=len(processes_to_resume))

    return {"count": len(processes_to_resume)}


@router.put("/{process_id}/abort", response_model=None, status_code=HTTPStatus.NO_CONTENT)
def abort_process_endpoint(process_id: UUID, request: Request, user: str = Depends(user_name)) -> None:
    process = _get_process(process_id)

    broadcast_func = api_broadcast_process_data(request)
    try:
        abort_process(process, user, broadcast_func=broadcast_func)
        broadcast_invalidate_status_counts()
        return
    except Exception as e:
        raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


@router.get("/status-counts", response_model=ProcessStatusCounts)
def status_counts() -> ProcessStatusCounts:
    """Retrieve status counts for processes and tasks."""

    stmt = (
        select(ProcessTable)
        .with_only_columns(ProcessTable.is_task, ProcessTable.last_status, count(ProcessTable.last_status))
        .group_by(ProcessTable.is_task, ProcessTable.last_status)
    )
    rows = db.session.execute(stmt).all()
    return ProcessStatusCounts(
        process_counts={status: num_processes for is_task, status, num_processes in rows if not is_task},
        task_counts={status: num_processes for is_task, status, num_processes in rows if is_task},
    )


@router.get("/{process_id}", response_model=ProcessSchema)
def show(process_id: UUID) -> dict[str, Any]:
    process = _get_process(process_id)
    p = load_process(process)

    return enrich_process(process, p)


def handle_process_error(message: str, **kwargs: Any) -> None:
    logger.debug(message, **kwargs)
    raise_status(HTTPStatus.BAD_REQUEST, message)


@trace
def _calculate_processes_crc32_checksum(results: list[ProcessTable]) -> int:
    """Calculate a CRC32 checksum of all the process id's and last_modified_at dates in order."""
    checksum = 0
    for p in results:
        checksum = zlib.crc32(p.process_id.bytes, checksum)
        last_modified_as_bytes = struct.pack("d", p.last_modified_at.timestamp())
        checksum = zlib.crc32(last_modified_as_bytes, checksum)
    return checksum


@router.get("/", response_model=list[ProcessSchema])
def processes_filterable(  # noqa: C901
    response: Response,
    range: str | None = None,
    sort: str | None = None,
    filter: str | None = None,
    if_none_match: str | None = Header(None),
) -> list[dict[str, Any]]:
    _range: list[int] | None = list(map(int, range.split(","))) if range else None
    _sort: list[str] | None = sort.split(",") if sort else None
    _filter: list[str] | None = filter.split(",") if filter else None

    # the joinedload on ProcessSubscriptionTable.subscription via ProcessBaseSchema.process_subscriptions prevents a query for every subscription later.
    # tracebacks are not presented in the list of processes and can be really large.
    processes: Select | CompoundSelect
    processes = select(ProcessTable).options(
        joinedload(ProcessTable.process_subscriptions)
        .joinedload(ProcessSubscriptionTable.subscription)
        .joinedload(SubscriptionTable.product),
        defer(ProcessTable.traceback),
    )

    if _filter is not None:
        if len(_filter) == 0 or (len(_filter) % 2) > 0:
            raise_status(HTTPStatus.BAD_REQUEST, "Invalid number of filter arguments")

        pydantic_filters = [Filter(field=field, value=value) for field, value in chunked(_filter, 2)]
        processes = filter_processes(processes, pydantic_filters, handle_process_error)

    if _sort is not None and len(_sort) >= 2:
        pydantic_sorting = [Sort(field=field, order=SortOrder[value.upper()]) for field, value in chunked(_sort, 2)]
        processes = sort_processes(processes, pydantic_sorting, handle_process_error)

    processes = add_response_range(processes, _range, response, unit="processes")

    results = list(db.session.scalars(processes).unique())

    # Calculate a CRC32 checksum of all the process id's and last_modified_at dates in order as entity tag
    checksum = _calculate_processes_crc32_checksum(results)

    entity_tag = hex(checksum)
    response.headers["ETag"] = f'W/"{entity_tag}"'

    # When the If-None-Match header contains the same CRC we can be sure that the resource has not changed,
    # so we can skip serialization at the backend and rerendering at the frontend.
    if if_none_match == entity_tag:
        raise CacheHit(HTTPStatus.NOT_MODIFIED, headers=dict(response.headers))

    return [enrich_process(p) for p in results]


ws_router = APIRouter()

if app_settings.ENABLE_WEBSOCKETS:

    @ws_router.websocket("/all/")
    async def websocket_process_list(websocket: WebSocket, token: str) -> None:
        error = await websocket_manager.authorize(websocket, token)

        await websocket.accept()
        if error:
            await websocket_manager.disconnect(websocket, reason=error)
            return

        channel = WS_CHANNELS.ALL_PROCESSES
        await websocket_manager.connect(websocket, channel)



from unittest import mock
from unittest.mock import Mock

from orchestrator.utils.get_subscription_dict import get_subscription_dict


@mock.patch("orchestrator.utils.get_subscription_dict._generate_etag")
async def test_get_subscription_dict_db(generate_etag, generic_subscription_1):
    generate_etag.side_effect = Mock(return_value="etag-mock")
    await get_subscription_dict(generic_subscription_1)
    assert generate_etag.called



import pytest
from orchestrator.forms import FormValidationError
from orchestrator.types import SubscriptionLifecycle

from products.product_types.example1 import Example1
from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example1_subscription):
    # given

    customer_id = "3f4fc287-0911-e511-80d0-005056956c1a"
    crm = CrmMocks(responses)
    crm.get_customer_by_uuid(customer_id)

    # TODO insert additional mocks, if needed (ImsMocks)

    # when

    init_state = {}

    result, process, step_log = run_workflow(
        "modify_example1",
        [{"subscription_id": example1_subscription}, init_state, {}],
    )

    # then

    assert_complete(result)
    state = extract_state(result)

    example1 = Example1.from_subscription(state["subscription_id"])
    assert example1.status == SubscriptionLifecycle.ACTIVE


@pytest.mark.workflow
def test_must_be_unused_to_change_mode(responses, example1_subscription):
    # given

    # TODO set test conditions or fixture so that "Mode can only be changed when there are no services attached to it" triggers

    # when

    init_state = {}

    with pytest.raises(FormValidationError) as error:
        result, _, _ = run_workflow("modify_example1", [{"subscription_id": example1_subscription}, init_state, {}])

    # then

    assert error.value.errors[0]["msg"] == "Mode can only be changed when there are no services attached to it"


@pytest.mark.workflow
def test_annotated_int_must_be_unique(responses, example1_subscription):
    # given

    # TODO set test conditions or fixture so that "annotated_int must be unique for example1" triggers

    # when

    init_state = {}

    with pytest.raises(FormValidationError) as error:
        result, _, _ = run_workflow("modify_example1", [{"subscription_id": example1_subscription}, init_state, {}])

    # then

    assert error.value.errors[0]["msg"] == "annotated_int must be unique for example1"



from uuid import uuid4

import pytest
from pydantic import conlist

from orchestrator.db import ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


@pytest.fixture
def test_product_type_list_union_overlap(test_product_block_one, test_product_sub_block_one):
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one

    def list_of_ports(t):
        return conlist(t, min_length=1)

    class ProductListUnionInactive(SubscriptionModel, is_base=True):
        test_block: ProductBlockOneForTestInactive | None
        list_union_blocks: list_of_ports(ProductBlockOneForTestInactive | SubBlockOneForTestInactive)

    class ProductListUnionProvisioning(ProductListUnionInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_block: ProductBlockOneForTestProvisioning
        list_union_blocks: list_of_ports(ProductBlockOneForTestProvisioning | SubBlockOneForTestProvisioning)

    class ProductListUnion(ProductListUnionProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockOneForTest
        list_union_blocks: list_of_ports(ProductBlockOneForTest | SubBlockOneForTest)

    SUBSCRIPTION_MODEL_REGISTRY["ProductListUnionOverlap"] = ProductListUnion
    yield ProductListUnionInactive, ProductListUnionProvisioning, ProductListUnion
    del SUBSCRIPTION_MODEL_REGISTRY["ProductListUnionOverlap"]


@pytest.fixture
def test_product_list_union_overlap(test_product_block_one_db):
    product = ProductTable(
        name="ProductListUnionOverlap",
        description="Test List Union Product Overlap",
        product_type="Test",
        tag="Union",
        status="active",
    )

    product_block, product_sub_block = test_product_block_one_db
    product.product_blocks = [product_block, product_sub_block]
    db.session.add(product)
    db.session.commit()
    return product.product_id


@pytest.fixture
def sub_list_union_overlap_subscription_1(
    test_product_list_union_overlap,
    test_product_type_list_union_overlap,
    test_product_sub_block_one,
    test_product_block_one,
    sub_one_subscription_1,
):
    ProductListUnionInactive, _, ProductListUnion = test_product_type_list_union_overlap
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    list_union_subscription_inactive = ProductListUnionInactive.from_product_id(
        product_id=test_product_list_union_overlap, customer_id=str(uuid4())
    )

    list_union_subscription_inactive.test_block = ProductBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id,
        int_field=3,
        str_field="",
        list_field=[1, 2],
        enum_field=DummyEnum.FOO,
        sub_block=SubBlockOneForTestInactive.new(
            subscription_id=list_union_subscription_inactive.subscription_id, int_field=3, str_field="2"
        ),
        sub_block_2=SubBlockOneForTestInactive.new(
            subscription_id=list_union_subscription_inactive.subscription_id, int_field=3, str_field="2"
        ),
        sub_block_list=[
            sub_one_subscription_1.test_block,
        ],
    )

    new_sub_block_1 = SubBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field=11, str_field="111"
    )
    new_sub_block_2 = SubBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field=12, str_field="121"
    )
    list_union_subscription_inactive.list_union_blocks = [new_sub_block_1, new_sub_block_2]
    list_union_subscription_inactive.save()

    list_union_subscription = ProductListUnion.from_other_lifecycle(
        list_union_subscription_inactive, SubscriptionLifecycle.ACTIVE
    )
    list_union_subscription.save()
    return list_union_subscription



from http import HTTPStatus
from ipaddress import IPv4Address
from unittest import mock
from uuid import uuid4

import pytest

from nwastdlib.url import URL
from orchestrator.api.helpers import product_block_paths
from orchestrator.db import (
    FixedInputTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import SubscriptionInstanceRelationTable, WorkflowTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import (
    RELATION_RESOURCE_TYPES,
    _generate_etag,
    build_extended_domain_model,
    get_subscription,
    unsync,
)
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus
from test.unit_tests.config import (
    IMS_CIRCUIT_ID,
    INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
    IPAM_PREFIX_ID,
    PARENT_IP_PREFIX_SUBSCRIPTION_ID,
    PEER_GROUP_SUBSCRIPTION_ID,
    PORT_SUBSCRIPTION_ID,
)
from test.unit_tests.conftest import do_refresh_subscriptions_search_view

SERVICE_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_BLOCK_ID = str(uuid4())
PROVISIONING_PORT_A_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_BLOCK_ID = str(uuid4())
IP_PREFIX_SUBSCRIPTION_ID = str(uuid4())
INVALID_SUBSCRIPTION_ID = str(uuid4())
INVALID_PORT_SUBSCRIPTION_ID = str(uuid4())

PRODUCT_ID = str(uuid4())
CUSTOMER_ID = str(uuid4())


@pytest.fixture
def seed():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
        ResourceTypeTable(resource_type=PORT_SUBSCRIPTION_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(PORT_A_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_ssp
    )
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(SSP_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_msp
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )
    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


# seeder to test direct relations using SubscriptionInstanceRelationTable instead of resource type.
@pytest.fixture
def seed_with_direct_relations():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=SSP_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_ssp_id = str(uuid4())
    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_ssp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_ssp,
    )
    lp_subscription_instance_ssp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_ssp_id,
        depends_on_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )

    lp_subscription_instance_msp_id = str(uuid4())
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_msp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_msp,
    )
    lp_subscription_instance_msp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_msp_id,
        depends_on_id=SSP_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()
    db.session.add(lp_subscription_instance_ssp_depends_on)
    db.session.add(lp_subscription_instance_msp_depends_on)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


def test_insync_404(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{str(uuid4())}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_insync_invalid_tagged(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{INVALID_PORT_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "reason": "subscription.not_in_sync",
        "create": [],
        "modify": [],
        "terminate": [],
        "system": [],
        "validate": [],
    }


def test_in_use_by_subscriptions_not_insync(seed, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = get_subscription(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_in_use_by_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_depends_on_subscriptions_not_insync(seed, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = db.session.get(SubscriptionTable, PORT_A_SUBSCRIPTION_ID, with_for_update=True)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_in_use_by_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = get_subscription(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_depends_on_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = get_subscription(PORT_A_SUBSCRIPTION_ID, for_update=True)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_delete_subscription_404(responses, seed, test_client):
    sub_id = uuid4()

    response = test_client.delete(f"/api/subscriptions/{sub_id}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_set_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(subscription_id)
    db.session.commit()

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.OK

    subscription = get_subscription(subscription_id)
    assert subscription.insync


def _create_failed_process(subscription_id):
    wf = WorkflowTable(
        workflow_id=uuid4(), name="validate_ip_prefix", description="validate_ip_prefix", target=Target.SYSTEM
    )
    process_id = uuid4()
    process = ProcessTable(
        process_id=process_id,
        workflow_id=wf.workflow_id,
        last_status=ProcessStatus.FAILED,
        last_step="Verify references in NSO",
        assignee="NOC",
        is_task=False,
    )
    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)
    db.session.add(wf)
    db.session.add(process)
    db.session.add(process_subscription)

    db.session.commit()


def test_try_set_failed_task_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(IP_PREFIX_SUBSCRIPTION_ID)
    db.session.commit()

    _create_failed_process(subscription_id)

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    subscription = get_subscription(subscription_id)
    assert not subscription.insync


def test_product_block_paths(generic_subscription_1, generic_subscription_2):
    subscription_1 = SubscriptionModel.from_subscription(generic_subscription_1)
    subscription_2 = SubscriptionModel.from_subscription(generic_subscription_2)
    assert product_block_paths(subscription_1) == ["product", "pb_1", "pb_2"]
    assert product_block_paths(subscription_2) == ["product", "pb_3"]


@pytest.mark.parametrize(
    "query, num_matches",
    [
        ("id", 7),
        ("tag:(POR* | LP)", 1),
        ("tag:SP", 3),
        ("tag:(POR* | LP) | tag:SP", 4),
        ("tag:(POR* | LP) | tag:SP -status:initial", 3),
    ],
)
def test_subscriptions_search(query, num_matches, seed, test_client, refresh_subscriptions_search_view):
    response = test_client.get(f"/api/subscriptions/search?query={query}")
    result = response.json()
    assert len(result) == num_matches


@pytest.fixture
def make_port_subscription():

    def make(subscription_id=None, customer_id=None, product_id=None):
        product_id = str(product_id or uuid4())
        subscription_id = str(subscription_id or uuid4())
        customer_id = str(customer_id or uuid4())
        product = ProductTable(
            product_id=product_id,
            name="ProductTable",
            description="description",
            product_type="Port",
            tag="Port",
            status="active",
        )
        subscription = SubscriptionTable(
            subscription_id=subscription_id,
            description="desc",
            status="active",
            insync=True,
            product=product,
            customer_id=customer_id,
        )

        db.session.add(product)
        db.session.add(subscription)
        db.session.commit()
        do_refresh_subscriptions_search_view()
        return subscription_id, customer_id, product_id

    return make


@pytest.mark.parametrize(
    "subscription_id",
    [
        "144cd85c-77a9-43bf-a78c-3ef2647c6ff1",
        "67bb12ae-8dfc-4ddb-9fd2-979392a9f2ad",
        "85ab4e94-c045-46f3-85c7-a8fbc7411aca",
        "27e84752-c700-491e-b126-454b5fa06dc0",
        "191b9852-e689-4dd2-8699-fc9c11c2bcdd",
        "998e7c12-e4d2-4321-a075-04b13a76ab39",
        "cfc64927-8f52-4325-87ac-daa1e16794e8",
        "5846383e-6387-411b-8e6e-bca022e2a86d",
    ],
)
def test_subscriptions_search_uuids2(
    subscription_id, make_port_subscription, test_client, refresh_subscriptions_search_view
):
    make_port_subscription(subscription_id=subscription_id)

    search_queries = {"full": subscription_id} | {
        f"group{n}": part for n, part in enumerate(subscription_id.split("-"))
    }

    def search(keyword):
        response = test_client.get(f"/api/subscriptions/search?query={keyword}")
        return len(response.json()) == 1

    results = {testcase: search(keyword) for testcase, keyword in search_queries.items()}

    succeeded = [search_queries[testcase] for testcase in results if results[testcase]]
    failed = [search_queries[testcase] for testcase in results if not results[testcase]]

    assert not failed, f"Could not find '{subscription_id}' by all keywords; {succeeded=} {failed=}"


def test_subscription_detail_with_domain_model(test_client, generic_subscription_1, benchmark, monitor_sqlalchemy):
    # test with a subscription that has domain model and without
    with monitor_sqlalchemy():

        @benchmark
        def response():
            return test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)

    assert response.status_code == HTTPStatus.OK
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_does_not_exist(test_client, generic_subscription_1, benchmark):
    # test with a subscription that has domain model and without
    @benchmark
    def response():
        return test_client.get(URL("api/subscriptions/domain-model") / uuid4())

    assert response.status_code == HTTPStatus.NOT_FOUND


def test_subscription_detail_with_domain_model_etag(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response.status_code == HTTPStatus.OK
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    assert etag == response.headers["ETag"]
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_if_none_match(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    response = test_client.get(
        URL("api/subscriptions/domain-model") / generic_subscription_1, headers={"If-None-Match": etag}
    )
    assert response.status_code == HTTPStatus.NOT_MODIFIED


def test_subscription_detail_with_in_use_by_ids_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(URL("api/subscriptions/domain-model") / product_one_subscription_1)
    assert response.status_code == HTTPStatus.OK
    assert not response.json()["block"]["sub_block"]["in_use_by_ids"]


@mock.patch("orchestrator.api.api_v1.endpoints.subscriptions.get_subscription_dict")
def test_subscription_detail_special_fields(mock_from_redis, test_client):
    """Test that a subscription with special field types is correctly serialized by Pydantic.

    https://github.com/pydantic/pydantic/issues/6669
    """
    standard_fields = {
        "subscription_id": "fabd6359-cb37-4a1c-bfc4-5c15aea7c888",
        "description": "desc",
        "status": "active",
        "customer_id": "f711c6fe-6de3-40bd-a4e7-9ac9d183a788",
        "insync": True,
        "version": 1,
        "product": {
            "name": "fake name",
            "description": "fake description",
            "product_type": "fake type",
            "status": "active",
            "tag": "fake tag",
        },
    }
    # Make the from_redis function return an IPv4Address - this wouldn't happen normally but it's easier
    # to mock than SubscriptionModel.from_subscription, and tests the special field formatting all the same
    special_fields = {"ip_address": IPv4Address("127.0.0.1")}
    mock_from_redis.return_value = (standard_fields | special_fields, "etag ofzo")

    response = test_client.get(URL("api/subscriptions/domain-model") / "fabd6359-cb37-4a1c-bfc4-5c15aea7c888")
    assert response.json() == {
        "subscription_id": "fabd6359-cb37-4a1c-bfc4-5c15aea7c888",
        "start_date": None,
        "description": "desc",
        "status": "active",
        "product_id": None,
        "customer_id": "f711c6fe-6de3-40bd-a4e7-9ac9d183a788",
        "insync": True,
        "note": None,
        "name": None,
        "end_date": None,
        "product": {
            "product_id": None,
            "name": "fake name",
            "description": "fake description",
            "product_type": "fake type",
            "status": "active",
            "tag": "fake tag",
            "created_at": None,
            "end_date": None,
        },
        "customer_descriptions": [],
        "tag": None,
        "version": 1,
        "ip_address": "127.0.0.1",
    }


def test_subscription_detail_with_in_use_by_ids_not_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(
        URL("api/subscriptions/domain-model") / product_one_subscription_1 / "?filter_owner_relations=false"
    )
    assert response.status_code == HTTPStatus.OK
    assert response.json()["block"]["sub_block"]["in_use_by_ids"]



import pytest

from orchestrator.api.helpers import getattr_in, product_block_paths, update_in


def test_product_block_paths(sub_list_union_overlap_subscription_1):
    paths = product_block_paths(sub_list_union_overlap_subscription_1)
    assert paths == [
        "product",
        "test_block.sub_block",
        "test_block.sub_block_2",
        "test_block.sub_block_list.0",
        "test_block",
        "list_union_blocks.0",
        "list_union_blocks.1",
    ]

    # Check that SubscriptionModel and subscription dict work the same
    assert product_block_paths(sub_list_union_overlap_subscription_1) == product_block_paths(
        sub_list_union_overlap_subscription_1.model_dump()
    )


class BasicObject:
    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)


SAMPLE_OBJECT = BasicObject(foo=BasicObject(bar="a", baz=[BasicObject(fooo=123)]))


@pytest.mark.parametrize(
    "input_object,path,expected_result",
    [
        ({"foo": "bar"}, "fizz", None),
        ({"foo": "bar"}, "foo", "bar"),
        ({"foo": {"bar": {"fizz": "buzz"}}}, "foo.bar.fizz", "buzz"),
        ({"foo": {"barbar": {"fizz": "buzz"}}}, "foo.bar.fizz", None),
        ({"foo": {"bar": {"fizz": ["a", "b", "c"]}}}, "foo.bar.fizz.1", "b"),
        ({"foo": {"bar": {"fizz": ["a", "b", "c"]}}}, "foo.bar.fizz.4", IndexError),
        ({"foo": {"bar": {"fizz": ["a", "b", "c"]}}}, "foo.bar", {"fizz": ["a", "b", "c"]}),
        (BasicObject(), "foo.bar", None),
        (SAMPLE_OBJECT, "foo.bar", "a"),
        (SAMPLE_OBJECT, "foo.barbar", None),
        (SAMPLE_OBJECT, "foo.baz.0.fooo", 123),
        (SAMPLE_OBJECT, "foo.baz.10.fooo", IndexError),
    ],
)
def test_getattr_in(input_object, path, expected_result):
    if isinstance(expected_result, type) and issubclass(expected_result, Exception):
        with pytest.raises(expected_result):
            getattr_in(input_object, path)
    else:
        assert getattr_in(input_object, path) == expected_result


@pytest.mark.parametrize(
    "input_dict,path,value,expected_result",
    [
        pytest.param({}, "foo", "bar", {"foo": "bar"}),
        pytest.param(
            {},
            "foo.x",
            "bar",
            {"foo": {"x": "bar"}},
            marks=[pytest.mark.xfail(reason="Intermediate dicts not created")],
        ),
        pytest.param({"foo": {}}, "foo.x", "bar", {"foo": {"x": "bar"}}),
        pytest.param(
            {},
            "foo.x.y",
            "bar",
            {"foo": {"x": {"y": "bar"}}},
            marks=[pytest.mark.xfail(reason="Intermediate dicts not created")],
        ),
        pytest.param({"foo": "bar"}, "fizz", "buzz", {"foo": "bar", "fizz": "buzz"}),
        pytest.param(
            {"foo": ["bar"]}, "foo.0", "buzz", {"foo": ["buzz"]}, marks=[pytest.mark.xfail(reason="Raises TypeError")]
        ),
    ],
)
def test_update_in(input_dict, path, value, expected_result):
    # TODO fix the failing scenarios
    assert update_in(input_dict, path, value) is None
    assert input_dict == expected_result



# Copyright 2019-2025 SURF, ESnet, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from shutil import copyfile

import jinja2
import typer
from alembic import command
from alembic.config import Config
from structlog import get_logger

import orchestrator.workflows
from orchestrator.cli.domain_gen_helpers.types import ModelUpdates
from orchestrator.cli.helpers.print_helpers import COLOR, str_fmt
from orchestrator.cli.migrate_domain_models import create_domain_models_migration_sql
from orchestrator.cli.migrate_tasks import create_tasks_migration_wizard
from orchestrator.cli.migrate_workflows import create_workflows_migration_wizard
from orchestrator.cli.migration_helpers import create_migration_file
from orchestrator.db import init_database
from orchestrator.settings import app_settings

logger = get_logger(__name__)

app: typer.Typer = typer.Typer()

orchestrator_module_location = os.path.dirname(orchestrator.__file__)
migration_dir = "migrations"

loader = jinja2.FileSystemLoader(os.path.join(orchestrator_module_location, f"{migration_dir}/templates"))
jinja_env = jinja2.Environment(
    loader=loader, autoescape=True, lstrip_blocks=True, trim_blocks=True, undefined=jinja2.StrictUndefined
)


def alembic_cfg() -> Config:
    cfg = Config("alembic.ini")
    version_locations = cfg.get_main_option("version_locations")
    cfg.set_main_option(
        "version_locations", f"{version_locations} {orchestrator_module_location}/{migration_dir}/versions/schema"
    )
    logger.info("Version Locations", locations=cfg.get_main_option("version_locations"))
    return cfg


@app.command(
    help="Initialize an empty migrations environment. This command will throw an exception when it detects conflicting files and directories."
)
def init() -> None:
    """Initialize the `migrations` directory.

    This command will initialize a migration directory for the orchestrator core application and setup a correct
    migration environment. It will also throw an exception when it detects conflicting files and directories.

    Returns:
        None

    CLI Options:
        None

    """

    if os.access(migration_dir, os.F_OK) and os.listdir(migration_dir):
        raise OSError(f"Directory {migration_dir} already exists and is not empty")

    logger.info("Creating directory", directory=os.path.abspath(migration_dir))
    os.makedirs(migration_dir)
    versions = os.path.join(migration_dir, "versions")
    logger.info("Creating directory", directory=os.path.abspath(versions))
    os.makedirs(versions)
    versions_schema = os.path.join(migration_dir, "versions/schema")
    logger.info("Creating directory", directory=os.path.abspath(versions_schema))
    os.makedirs(versions_schema)

    source_env_py = os.path.join(orchestrator_module_location, f"{migration_dir}/templates/env.py.j2")
    env_py = os.path.join(migration_dir, "env.py")
    logger.info("Creating file", file=os.path.abspath(env_py))
    copyfile(source_env_py, env_py)

    source_script_py_mako = os.path.join(orchestrator_module_location, f"{migration_dir}/script.py.mako")
    script_py_mako = os.path.join(migration_dir, "script.py.mako")
    logger.info("Creating file", file=os.path.abspath(script_py_mako))
    copyfile(source_script_py_mako, script_py_mako)

    source_helpers_py = os.path.join(orchestrator_module_location, f"{migration_dir}/templates/helpers.py.j2")
    helpers_py = os.path.join(migration_dir, "helpers.py")
    logger.info("Creating file", file=os.path.abspath(helpers_py))
    copyfile(source_helpers_py, helpers_py)

    template = jinja_env.get_template("alembic.ini.j2")

    if not os.access(os.path.join(os.getcwd(), "alembic.ini"), os.F_OK):
        logger.info("Creating file", file=os.path.join(os.getcwd(), "alembic.ini"))
        with open(os.path.join(os.getcwd(), "alembic.ini"), "w") as alembic_ini:
            alembic_ini.write(template.render(migrations_dir=migration_dir))
    else:
        logger.info("Skipping Alembic.ini file. It already exists")


@app.command(help="Get the database heads")
def heads() -> None:
    """The `heads` command command shows the Alembic database heads.

    CLI Options:
        None

    """
    command.heads(alembic_cfg())


@app.command(help="Merge database revisions.")
def merge(
    revisions: str = typer.Argument(default=None, help="Add the revision you would like to merge to this command."),
    message: str = typer.Option(None, "--message", "-m", help="The revision message"),
) -> None:
    """Merge database revisions.

    It is possible when using multiple git branches in your WFO development lifecycle to have
    multiple Alembic heads emerge. This command will allow you to merge those two (or more)
    heads to resolve the issue. You also might need to run this after updating your version
    of orchestrator-core if there have been schema changes.

    [Read More Here](https://alembic.sqlalchemy.org/en/latest/branches.html#merging-branches)

    Args:
        revisions: List of revisions to merge
        message: Optional message for the revision.

    Returns:
        None

    CLI Options:
        ```sh
        Arguments:
            [REVISIONS]  Add the revision you would like to merge to this command.

        Options:
            -m, --message TEXT  The revision message
        ```
    """
    command.merge(alembic_cfg(), revisions, message=message)


@app.command()
def upgrade(revision: str = typer.Argument(help="Rev id to upgrade to")) -> None:
    """The `upgrade` command will upgrade the database to the specified revision.

    Args:
        revision: Optional argument to indicate where to upgrade to.

    Returns:
        None

    CLI Options:
        ```sh
        Arguments:
            [REVISION]  Rev id to upgrade to

        Options:
            --help  Show this message and exit.
        ```

    """
    command.upgrade(alembic_cfg(), revision)


@app.command()
def downgrade(revision: str = typer.Argument("-1", help="Rev id to downgrade to")) -> None:
    """The `downgrade` command will downgrade the database to the previous revision or to the optionally specified revision.

    Args:
        revision (str, optional): Optional argument to indicate where to downgrade to. [default: -1]

    Returns:
        None

    CLI Options:
        ```sh
        Arguments:
            [REVISION]  Rev id to upgrade to  [default: -1]
        ```

    """
    command.downgrade(alembic_cfg(), revision)


@app.command()
def revision(
    message: str = typer.Option(None, "--message", "-m", help="The revision message"),
    version_path: str = typer.Option(None, "--version-path", help="Specify specific path from config for version file"),
    autogenerate: bool = typer.Option(False, help="Detect schema changes and add migrations"),
    head: str = typer.Option("data@head", help="Determine the head you need to add your migration to."),
) -> None:
    """The `revision` command creates a new Alembic revision file.

    Args:
        message: The revision message
        version_path: Specify specific path from config for version file
        autogenerate: Whether to detect schema changes.
        head: To which head the migration applies

    Returns:
        None

    CLI Options:
        ```sh
        Options:
            -m, --message TEXT              The revision message
            --version-path TEXT             Specify specific path from config for version file
            --autogenerate / --no-autogenerate
                                            Detect schema changes and add migrations [default: no-autogenerate]
            --head TEXT                     Determine the head you need to add your migration to.
        ```
    """
    command.revision(alembic_cfg(), message, version_path=version_path, autogenerate=autogenerate, head=head)


@app.command()
def history(
    verbose: bool = typer.Option(False, "--verbose", "-v", help="Verbose output"),
    indicate_current: bool = typer.Option(True, "--current", "-c", help="Indicate current revision"),
) -> None:
    """The `history` command lists Alembic revision history/changeset scripts in chronological order.

    Args:
        verbose (bool, optional): Verbose output
        indicate_current (bool, optional): Indicate current revision

    Returns:
        None

    CLI Options:
        ```sh
        Options:
            -v, --verbose  Verbose output
            -c, --current  Indicate current revision
        ```
    """
    command.history(alembic_cfg(), verbose=verbose, indicate_current=indicate_current)


@app.command(help="Create revision based on diff_product_in_database.")
def migrate_domain_models(
    message: str = typer.Argument(..., help="Migration name"),
    test: bool = typer.Option(False, help="Optional boolean if you don't want to generate a migration file"),
    inputs: str = typer.Option("{}", help="Stringified dict to prefill inputs"),
    updates: str = typer.Option("{}", help="Stringified dict to map updates instead of using inputs"),
) -> tuple[list[str], list[str]] | None:
    """Create migration file based on SubscriptionModel.diff_product_in_database. BACKUP DATABASE BEFORE USING THE MIGRATION!.

    You will be prompted with inputs for new models and resource type updates.
    Resource type updates are only handled when it's renamed in all product blocks.

    Args:
        message: Message/description of the generated migration.
        test: Optional boolean if you don't want to generate a migration file.
        inputs: stringified dict to prefill inputs.
            The inputs and updates argument is mostly used for testing, prefilling the given inputs, here examples:
            - new product: `inputs = { "new_product_name": { "description": "add description", "product_type": "add_type", "tag": "add_tag" }}`
            - new product fixed input: `inputs = { "new_product_name": { "new_fixed_input_name": "value" }}`
            - new product block: `inputs = { "new_product_block_name": { "description": "add description", "tag": "add_tag" } }`
            - new resource type: `inputs = { "new_resource_type_name": { "description": "add description", "value": "add default value", "new_product_block_name": "add default value for block" }}`
                - `new_product_block_name` prop inserts value specifically for that block.
                - `value` prop is inserted as default for all existing instances it is added to.

        updates: stringified dict to prefill inputs.
            - renaming a fixed input:
                - `updates = { "fixed_inputs": { "product_name": { "old_fixed_input_name": "new_fixed_input_name" } } }`
            - renaming a resource type to a new resource type:
                - `inputs = { "new_resource_type_name": { "description": "add description" }}`
                - `updates = { "resource_types": { "old_resource_type_name": "new_resource_type_name" } }`
            - renaming a resource type to existing resource type: `updates = { "resource_types": { "old_resource_type_name": "new_resource_type_name" } }`

    Returns None unless `--test` is used, in which case it returns:
        - tuple:
            - list of upgrade SQL statements in string format.
            - list of downgrade SQL statements in string format.
    """
    if not app_settings.TESTING:
        init_database(app_settings)

    if test:
        print(  # noqa: T001, T201
            f"{str_fmt('NOTE:', flags=[COLOR.BOLD, COLOR.CYAN])} Running in test mode. No migration file will be generated.\n"
        )

    inputs_dict = json.loads(inputs) if isinstance(inputs, str) else {}
    updates_dict = json.loads(updates) if isinstance(updates, str) else {}
    updates_class = None
    if updates_dict:
        updates_class = ModelUpdates(
            fixed_inputs=updates_dict.get("fixed_inputs", {}),
            resource_types=updates_dict.get("resource_types", {}),
            block_resource_types=updates_dict.get("block_resource_types", {}),
        )
    sql_upgrade_stmts, sql_downgrade_stmts = create_domain_models_migration_sql(inputs_dict, updates_class, bool(test))

    if test:
        return sql_upgrade_stmts, sql_downgrade_stmts

    sql_upgrade_str = "\n".join(
        [f'    conn.execute(sa.text("""\n{sql_stmt}\n    """))' for sql_stmt in sql_upgrade_stmts]
    )
    sql_downgrade_str = "\n".join(
        [f'    conn.execute(sa.text("""\n{sql_stmt}\n    """))' for sql_stmt in sql_downgrade_stmts]
    )
    create_migration_file(alembic_cfg(), sql_upgrade_str, sql_downgrade_str, message)
    return None


@app.command(help="Create migration file based on diff workflows in db")
def migrate_workflows(
    message: str = typer.Argument(..., help="Migration name"),
    test: bool = typer.Option(False, help="Optional boolean if you don't want to generate a migration file"),
) -> tuple[list[dict], list[dict]] | None:
    """The `migrate-workflows` command creates a migration file based on the difference between workflows in the database and registered WorkflowInstances in your codebase.

    !!! warning "BACKUP YOUR DATABASE BEFORE USING THE MIGRATION!"

    You will be prompted with inputs for new models and resource type updates.
    Resource type updates are only handled when it's renamed in all product blocks.

    Args:
        message: Message/description of the generated migration.
        test: Optional boolean if you don't want to generate a migration file.

    Returns None unless `--test` is used, in which case it returns:
        - tuple:
            - list of upgrade SQL statements in string format.
            - list of downgrade SQL statements in string format.

    CLI Arguments:
        ```sh
        Arguments:
            MESSAGE  Migration name  [required]

        Options:
            --test / --no-test  Optional boolean if you don't want to generate a migration
            file  [default: no-test]
        ```
    """
    if not app_settings.TESTING:
        init_database(app_settings)

    if test:
        print(  # noqa: T001, T201
            f"{str_fmt('NOTE:', flags=[COLOR.BOLD, COLOR.CYAN])} Running in test mode. No migration file will be generated.\n"
        )

    workflows_to_add, workflows_to_delete = create_workflows_migration_wizard()

    # String 'template' arguments
    import_str = "from orchestrator.migrations.helpers import create_workflow, delete_workflow\n"
    tpl_preamble_lines = []
    tpl_upgrade_lines = []
    tpl_downgrade_lines = []

    if workflows_to_add:
        tpl_preamble_lines.append(f"new_workflows = {json.dumps(workflows_to_add, indent=4)}\n")
        tpl_upgrade_lines.extend(
            [(" " * 4) + "for workflow in new_workflows:", (" " * 8) + "create_workflow(conn, workflow)"]
        )
        tpl_downgrade_lines.extend(
            [(" " * 4) + "for workflow in new_workflows:", (" " * 8) + 'delete_workflow(conn, workflow["name"])']
        )

    if workflows_to_delete:
        tpl_preamble_lines.append(f"old_workflows = {json.dumps(workflows_to_delete, indent=4)}\n")
        tpl_upgrade_lines.extend(
            [(" " * 4) + "for workflow in old_workflows:", (" " * 8) + 'delete_workflow(conn, workflow["name"])']
        )
        tpl_downgrade_lines.extend(
            [(" " * 4) + "for workflow in old_workflows:", (" " * 8) + "create_workflow(conn, workflow)"]
        )

    preamble = "\n".join(
        [
            import_str,
            *tpl_preamble_lines,
        ]
    )
    sql_upgrade_str = "\n".join(tpl_upgrade_lines)
    sql_downgrade_str = "\n".join(tpl_downgrade_lines)

    if test:
        return workflows_to_add, workflows_to_delete

    create_migration_file(alembic_cfg(), sql_upgrade_str, sql_downgrade_str, message, preamble=preamble)
    return None


@app.command(help="Create migration file based on diff tasks in db")
def migrate_tasks(
    message: str = typer.Argument(..., help="Migration name"),
    test: bool = typer.Option(False, help="Optional boolean if you don't want to generate a migration file"),
) -> tuple[list[dict], list[dict]] | None:
    """The `migrate-tasks` command creates a migration file based on the difference between tasks in the database and registered TaskInstances in your codebase.

    !!! warning "BACKUP YOUR DATABASE BEFORE USING THE MIGRATION!"

    You will be prompted with inputs for new models and resource type updates.
    Resource type updates are only handled when it's renamed in all product blocks.

    Args:
        message: Message/description of the generated migration.
        test: Optional boolean if you don't want to generate a migration file.

    Returns None unless `--test` is used, in which case it returns:
        - tuple:
            - list of upgrade SQL statements in string format.
            - list of downgrade SQL statements in string format.

    CLI Arguments:
        ```sh
        Arguments:
            MESSAGE  Migration name  [required]

        Options:
            --test / --no-test  Optional boolean if you don't want to generate a migration
            file  [default: no-test]
        ```
    """
    if not app_settings.TESTING:
        init_database(app_settings)

    if test:
        print(  # noqa: T001, T201
            f"{str_fmt('NOTE:', flags=[COLOR.BOLD, COLOR.CYAN])} Running in test mode. No migration file will be generated.\n"
        )

    tasks_to_add, tasks_to_delete = create_tasks_migration_wizard()

    # String 'template' arguments
    import_str = "from orchestrator.migrations.helpers import create_task, delete_workflow\n"
    tpl_preamble_lines = []
    tpl_upgrade_lines = []
    tpl_downgrade_lines = []

    if tasks_to_add:
        tpl_preamble_lines.append(f"new_tasks = {json.dumps(tasks_to_add, indent=4)}\n")
        tpl_upgrade_lines.extend([(" " * 4) + "for task in new_tasks:", (" " * 8) + "create_task(conn, task)"])
        tpl_downgrade_lines.extend(
            [(" " * 4) + "for task in new_tasks:", (" " * 8) + 'delete_workflow(conn, task["name"])']
        )

    if tasks_to_delete:
        tpl_preamble_lines.append(f"old_tasks = {json.dumps(tasks_to_delete, indent=4)}\n")
        tpl_upgrade_lines.extend(
            [(" " * 4) + "for task in old_tasks:", (" " * 8) + 'delete_workflow(conn, task["name"])']
        )
        tpl_downgrade_lines.extend([(" " * 4) + "for task in old_tasks:", (" " * 8) + "create_task(conn, task)"])

    preamble = "\n".join(
        [
            import_str,
            *tpl_preamble_lines,
        ]
    )
    sql_upgrade_str = "\n".join(tpl_upgrade_lines)
    sql_downgrade_str = "\n".join(tpl_downgrade_lines)

    if test:
        return tasks_to_add, tasks_to_delete

    create_migration_file(alembic_cfg(), sql_upgrade_str, sql_downgrade_str, message, preamble=preamble)
    return None



from unittest import mock

import pytest
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound

from orchestrator.db import ResourceTypeTable, SubscriptionTable, WorkflowTable, db, transactional
from orchestrator.db.helpers import get_postgres_version
from orchestrator.targets import Target


def test_transactional():
    def insert_wf(state):
        wf = WorkflowTable(name="Test transactional", target=Target.CREATE, description="Testing 1, 2, 3!")
        db.session.add(wf)

    def insert_wf_error(state):
        wf = WorkflowTable(
            name="Test transactional [ERROR]", target=Target.CREATE, description="Testing 1, 2, 3! BOOM!"
        )
        db.session.add(wf)
        raise Exception("Let's wreck some havoc!")

    logger = mock.MagicMock()

    with transactional(db, logger):
        insert_wf({})
    logger.assert_has_calls(
        [
            mock.call.debug("Temporarily disabling commit."),
            mock.call.debug("Reenabling commit."),
            mock.call.debug("Committing transaction."),
        ]
    )

    logger.reset_mock()
    with pytest.raises(Exception) as ex:
        with transactional(db, logger):
            insert_wf_error({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Let's wreck some havoc!"

    logger.assert_has_calls(
        [
            mock.call.debug("Temporarily disabling commit."),
            mock.call.debug("Reenabling commit."),
            mock.call.warning("Rolling back transaction."),
        ]
    )


def test_transactional_no_commit():
    def insert_wf(state):
        wf = WorkflowTable(
            name="Test transactional should not be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
        )
        db.session.add(wf)
        db.session.commit()

        raise Exception("Lets rollback")

    logger = mock.MagicMock()

    with pytest.raises(Exception, match="Lets rollback") as ex:
        with transactional(db, logger):
            insert_wf({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Lets rollback"

    assert (
        db.session.scalars(
            select(WorkflowTable).filter(WorkflowTable.name == "Test transactional should not be committed")
        ).all()
        == []
    )
    logger.assert_has_calls(
        [
            mock.call.warning(
                "Step function tried to issue a commit. It should not! Will execute commit on behalf of step function when it returns."
            ),
        ]
    )


def test_transactional_no_commit_second_thread():
    def insert_wf(state):
        wf = WorkflowTable(
            name="Test transactional should not be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
        )
        db.session.add(wf)
        db.session.commit()

        # Create new database session to simulate another workflow/api handler running at the same time
        # This is also a workaround for our disable commit wrapper, but it should be reasonable obvious that
        # someone is fucking around if you see `with db.database_scope():` in actual production code

        with db.database_scope():
            wf2 = WorkflowTable(
                name="Test transactional should be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
            )
            db.session.add(wf2)
            db.session.commit()

        raise Exception("Lets rollback")

    logger = mock.MagicMock()

    with pytest.raises(Exception, match="Lets rollback") as ex:
        with transactional(db, logger):
            insert_wf({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Lets rollback"

    assert (
        db.session.execute(
            select(WorkflowTable).filter(WorkflowTable.name == "Test transactional should not be committed")
        ).all()
        == []
    )
    logger.assert_has_calls(
        [
            mock.call.warning(
                "Step function tried to issue a commit. It should not! Will execute commit on behalf of step function when it returns."
            ),
        ]
    )


def test_autouse_fixture_rolls_back_aaa():
    # We want to test whether a change committed to the database in one test is visible to other tests (as in really
    # persisted to the database). Of course such a change should not be visible if our `fastapi_app` and `database`
    # autouse fixtures work as advertised.
    #
    # However, tests should be independent of each other and we cannot assume one test runs before the other. Hence
    # this test comes in two versions: one with the `_aaa` postfix and one with the `_bbb` postfix. Both will test
    # for the presence of a change the other test thinks it has committed to the database. If one of the tests (the
    # one that runs after the other) finds the change the other has committed our fixtures don't work properly.

    # Using ResourceTypeTable as it's a simple model than doesn't require foreign keys.
    rt = ResourceTypeTable(resource_type="aaa", description="aaa")
    # print(db)
    # print(dir(db))

    db.session.add(rt)
    db.session.commit()

    with pytest.raises(NoResultFound):
        db.session.scalars(select(ResourceTypeTable).where(ResourceTypeTable.resource_type == "bbb")).one()


def test_autouse_fixture_rolls_back_bbb():
    # We want to test whether a change committed to the database in one test is visible to other tests (as in really
    # persisted to the database). Of course such a change should not be visible if our `fastapi_app` and `database`
    # autouse fixtures work as advertised.
    #
    # However, tests should be independent of each other and we cannot assume one test runs before the other. Hence
    # this test comes in two versions: one with the `_aaa` postfix and one with the `_bbb` postfix. Both will test
    # for the presence of a change the other test thinks it has committed to the database. If one of the tests (the
    # one that runs after the other) finds the change the other has committed our fixtures don't work properly.

    # Using ResourceTypeTable as it's a simple model than doesn't require foreign keys.
    rt = ResourceTypeTable(resource_type="bbb", description="bbb")
    db.session.add(rt)
    db.session.commit()

    with pytest.raises(NoResultFound):
        db.session.scalars(select(ResourceTypeTable).where(ResourceTypeTable.resource_type == "aaa")).one()


def test_str_method():
    assert (
        str(SubscriptionTable())
        == "SubscriptionTable(subscription_id=None, description=None, status=None, product_id=None, customer_id=None, insync=None, start_date=None, end_date=None, note=None, version=None)"
    )


def test_get_postgres_version():
    pg_version = get_postgres_version()
    assert isinstance(pg_version, int)
    assert 13 < pg_version < 20



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import singledispatch
from http import HTTPStatus
from typing import Any, cast

import structlog

from nwastdlib.ex import show_ex
from orchestrator.types import ErrorDict
from pydantic_forms.types import JSON

logger = structlog.get_logger(__name__)


class ApiException(Exception):  # noqa: N818
    """Api Exception Class.

    This is a copy of what is generated in api_clients. We use this to have consistent error handling for nso to.
    This should conform to what is used in the api clients.
    """

    status: HTTPStatus | None
    reason: str | None
    body: str | None
    headers: dict[str, str]

    def __init__(self, status: HTTPStatus | None = None, reason: str | None = None, http_resp: object | None = None):
        super().__init__(status, reason, http_resp)
        if http_resp:
            self.status = http_resp.status  # type:ignore
            self.reason = http_resp.reason  # type:ignore
            self.body = http_resp.data  # type:ignore
            self.headers = http_resp.getheaders()  # type:ignore
        else:
            self.status = status
            self.reason = reason
            self.body = None
            self.headers = {}

    def __str__(self) -> str:
        """Create custom error messages for exception."""
        error_message = "({})\n" "Reason: {}\n".format(self.status, self.reason)
        if self.headers:
            error_message += f"HTTP response headers: {self.headers}\n"

        if self.body:
            error_message += f"HTTP response body: {self.body}\n"

        return error_message


class ProcessFailureError(Exception):
    message: str
    details: JSON

    def __init__(self, message: str, details: JSON = None) -> None:
        super().__init__(message, details)
        self.message = message
        self.details = details


class InconsistentDataError(ProcessFailureError):
    pass


class StaleDataError(ValueError):
    """The version of the update payload does not match the version in the database."""

    def __init__(self, current_version: int, new_version: int | None = None) -> None:
        message = f"Stale data: given version ({new_version}) does not match the current version ({current_version})"
        super().__init__(message)


def is_api_exception(ex: Exception) -> bool:
    """Test for swagger-codegen ApiException.

    For each API, swagger-codegen generates a new ApiException class. These are not organized into
    a hierarchy. Hence, testing whether one is dealing with one of the ApiException classes without knowing how
    many there are and where they are located, needs some special logic.

    Args:
        ex: the Exception to be tested.

    Returns:
        True if it is an ApiException, False otherwise.

    """
    return ex.__class__.__name__ == "ApiException"


@singledispatch
def error_state_to_dict(err: Any) -> ErrorDict:
    """Return an ErrorDict based on the passed error object.

    Args:
        err: An error object like an Exception, Error or ErrorDict
    Returns:
        An ErrorDict containing the error message a status_code and a traceback if available

    """
    raise NotImplementedError(f"Unsupported error state type: {type(err)}")


@error_state_to_dict.register(dict)
def _(err: ErrorDict) -> ErrorDict:
    return err


@error_state_to_dict.register
def _(err: ProcessFailureError) -> ErrorDict:
    return {"class": type(err).__name__, "error": err.message, "traceback": show_ex(err), "details": err.details}


@error_state_to_dict.register
def _(err: Exception) -> ErrorDict:
    # We can't dispatch on ApiException, see is_api_exception docstring
    if is_api_exception(err):
        err = cast(ApiException, err)
        return {
            "class": type(err).__name__,
            "error": err.reason,
            "status_code": err.status,
            "body": err.body,
            "headers": "\n".join(f"{k}: {v}" for k, v in err.headers.items()),
            "traceback": show_ex(err),
        }

    return {"class": type(err).__name__, "error": str(err), "traceback": show_ex(err)}



from orchestrator.cli.generator.generator.enums import convert_int_enum, convert_str_enum


def test_convert_str_enum():
    field = {
        "default": "tagged",
        "enum_type": "str",
        "name": "port_mode",
        "type": "enum",
        "values": ["untagged", "tagged", "link_member"],
    }
    converted_field = convert_str_enum(field)
    assert converted_field["type"] == "PortMode"
    assert converted_field["default"] == "PortMode.tagged"


def test_convert_int_enum():
    field = {
        "default": 40000,
        "enum_type": "int",
        "name": "speed",
        "type": "enum",
        "values": [1000, 10000, 40000, 100000],
    }
    converted_field = convert_int_enum(field)
    assert converted_field["type"] == "Speed"
    assert converted_field["default"] == "Speed._40000"



from orchestrator.db import db
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscription_relations import get_depends_on_subscriptions, get_in_use_by_subscriptions
from orchestrator.types import SubscriptionLifecycle


def terminate_subscription(subscription_id):
    subscription = SubscriptionModel.from_subscription(subscription_id)
    terminated_subscription = SubscriptionModel.from_other_lifecycle(subscription, SubscriptionLifecycle.TERMINATED)
    terminated_subscription.save()
    db.session.commit()
    return terminated_subscription


async def test_get_in_use_by_subscriptions(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [sub_one_subscription_1.subscription_id, sub_two_subscription_1.subscription_id]

    terminate_subscription(product_sub_list_union_subscription_1)
    terminate_subscription(sub_two_subscription_1.subscription_id)

    result = await get_in_use_by_subscriptions(subscription_ids, ())

    # then
    expected_result = [
        # sub_one_subscription_1 in_use_by_subscriptions
        sorted([product_sub_list_union_subscription_1, sub_list_union_overlap_subscription_1.subscription_id]),
        # sub_two_subscription_1 in_use_by_subscriptions
        [product_sub_list_union_subscription_1],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_in_use_by_subscriptions_only_active(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [sub_one_subscription_1.subscription_id, sub_two_subscription_1.subscription_id]

    terminate_subscription(product_sub_list_union_subscription_1)
    terminate_subscription(sub_two_subscription_1.subscription_id)

    result = await get_in_use_by_subscriptions(subscription_ids, ("active",))

    # then
    expected_result = [
        # sub_one_subscription_1 in_use_by_subscriptions
        sorted(
            [
                sub_list_union_overlap_subscription_1.subscription_id,  # ACTIVE
            ]
        ),
        # sub_two_subscription_1 in_use_by_subscriptions - terminated
        [],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_in_use_by_subscriptions_only_terminated(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [sub_one_subscription_1.subscription_id, sub_two_subscription_1.subscription_id]

    terminate_subscription(product_sub_list_union_subscription_1)
    terminate_subscription(sub_two_subscription_1.subscription_id)

    result = await get_in_use_by_subscriptions(subscription_ids, ("terminated",))

    # then
    expected_result = [
        # sub_one_subscription_1 in_use_by_subscriptions
        [product_sub_list_union_subscription_1],
        # sub_two_subscription_1 in_use_by_subscriptions - terminated
        [product_sub_list_union_subscription_1],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_in_use_by_subscriptions_empty():
    # when
    subscription_ids = []

    result = await get_in_use_by_subscriptions(subscription_ids, ())

    # then
    expected_result = []

    assert result == expected_result


async def test_get_depends_on_subscriptions(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [product_sub_list_union_subscription_1, sub_list_union_overlap_subscription_1.subscription_id]

    result = await get_depends_on_subscriptions(subscription_ids, ())

    # then
    expected_result = [
        # product_sub_list_union_subscription_1 in_use_by_subscriptions
        sorted([sub_two_subscription_1.subscription_id, sub_one_subscription_1.subscription_id]),
        # sub_list_union_overlap_subscription_1 in_use_by_subscriptions
        [sub_one_subscription_1.subscription_id],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_depends_on_subscriptions_only_active(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [product_sub_list_union_subscription_1, sub_list_union_overlap_subscription_1.subscription_id]
    terminate_subscription(product_sub_list_union_subscription_1)
    terminate_subscription(sub_two_subscription_1.subscription_id)

    result = await get_depends_on_subscriptions(subscription_ids, ("active",))

    # then
    expected_result = [
        # product_sub_list_union_subscription_1 in_use_by_subscriptions - terminated
        [sub_one_subscription_1.subscription_id],
        # sub_list_union_overlap_subscription_1 in_use_by_subscriptions
        [sub_one_subscription_1.subscription_id],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_depends_on_subscriptions_only_terminated(
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    sub_list_union_overlap_subscription_1,
):
    # when
    subscription_ids = [product_sub_list_union_subscription_1, sub_list_union_overlap_subscription_1.subscription_id]
    terminate_subscription(product_sub_list_union_subscription_1)
    terminate_subscription(sub_two_subscription_1.subscription_id)

    result = await get_depends_on_subscriptions(subscription_ids, ("terminated",))

    # then
    expected_result = [
        # product_sub_list_union_subscription_1 in_use_by_subscriptions - terminated
        sorted(
            [
                sub_two_subscription_1.subscription_id,  # TERMINATED
                # sub_one_subscription_1.subscription_id # ACTIVE
            ]
        ),
        # sub_list_union_overlap_subscription_1 in_use_by_subscriptions sub_one_subscription_1
        [],
    ]

    assert [sorted([sub.subscription_id for sub in r_list]) for r_list in result] == expected_result


async def test_get_depends_on_subscriptions_empty():
    # when
    subscription_ids = []

    result = await get_depends_on_subscriptions(subscription_ids, ())

    # then
    expected_result = []

    assert result == expected_result



from http import HTTPStatus
from uuid import uuid4

import pytest

from orchestrator.db import ProductTable, SubscriptionCustomerDescriptionTable, SubscriptionTable, db

SUBSCRIPTION_ID = uuid4()
SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID = uuid4()
CUSTOMER_ID = str(uuid4())


@pytest.fixture
def seed():
    product = ProductTable(
        name="ProductTable", description="description", product_type="Port", tag="Port", status="active"
    )
    subscription = SubscriptionTable(
        subscription_id=SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=product,
        customer_id=str(uuid4()),
        customer_descriptions=[
            SubscriptionCustomerDescriptionTable(
                id=SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID, customer_id=CUSTOMER_ID, description="customer other alias"
            )
        ],
        version=1,
    )

    db.session.add(product)
    db.session.add(subscription)
    db.session.commit()


def test_get_by_id(seed, test_client):
    response = test_client.get(f"/api/subscription_customer_descriptions/{SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID}")
    assert HTTPStatus.OK == response.status_code


def test_get_404(seed, test_client):
    random_uuid = uuid4()
    response = test_client.get(f"/api/subscription_customer_descriptions/{random_uuid}")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_get_by_customer_and_subscription_id(seed, test_client):
    url = f"/api/subscription_customer_descriptions/customer/{CUSTOMER_ID}/subscription/{SUBSCRIPTION_ID}"
    response = test_client.get(url)
    assert HTTPStatus.OK == response.status_code
    assert response.json()["id"] == str(SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID)


def test_save(seed, test_client):
    body = {
        "subscription_id": SUBSCRIPTION_ID,
        "customer_id": str(uuid4()),
        "description": "Customer specific alias",
    }
    response = test_client.post("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response.status_code

    count = db.session.query(SubscriptionCustomerDescriptionTable).count()
    assert 2 == count


def test_update(seed, test_client):
    new_desc = "Updated specific alias"
    body = {
        "id": SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID,
        "subscription_id": SUBSCRIPTION_ID,
        "customer_id": CUSTOMER_ID,
        "description": new_desc,
    }
    response = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response.status_code

    count = db.session.query(SubscriptionCustomerDescriptionTable).count()
    assert 1 == count

    customer_description = db.session.query(SubscriptionCustomerDescriptionTable).first()
    assert new_desc == customer_description.description
    assert 2 == customer_description.version


def test_update_with_version(seed, test_client):
    new_desc = "Updated specific alias"
    version = 1
    body = {
        "id": SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID,
        "subscription_id": SUBSCRIPTION_ID,
        "customer_id": CUSTOMER_ID,
        "description": new_desc,
        "version": version,
    }
    response = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response.status_code

    count = db.session.query(SubscriptionCustomerDescriptionTable).count()
    assert 1 == count

    customer_description = db.session.query(SubscriptionCustomerDescriptionTable).first()
    assert new_desc == customer_description.description
    assert version + 1 == customer_description.version


def test_update_with_lower_version_invalid(seed, test_client):
    new_desc = "Updated specific alias"
    body = {
        "id": SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID,
        "subscription_id": SUBSCRIPTION_ID,
        "customer_id": CUSTOMER_ID,
        "description": new_desc,
        "version": 0,
    }
    response = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.BAD_REQUEST == response.status_code

    data = response.json()
    assert "Stale data: given version (0) does not match the current version (1)" == data["detail"]


def test_update_with_higher_version_invalid(seed, test_client):
    new_desc = "Updated specific alias"
    body = {
        "id": SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID,
        "subscription_id": SUBSCRIPTION_ID,
        "customer_id": CUSTOMER_ID,
        "description": new_desc,
        "version": 10,
    }
    response = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.BAD_REQUEST == response.status_code

    data = response.json()
    assert "Stale data: given version (10) does not match the current version (1)" == data["detail"]


def test_delete(seed, test_client):
    test_client.delete(f"/api/subscription_customer_descriptions/{SUBSCRIPTION_CUSTOMER_DESCRIPTION_ID}")

    count = db.session.query(SubscriptionCustomerDescriptionTable).count()
    assert 0 == count



from collections.abc import Generator
from itertools import chain

from sqlalchemy import select
from sqlalchemy.sql.expression import Delete, Insert
from sqlalchemy.sql.selectable import ScalarSelect

from orchestrator.cli.domain_gen_helpers.helpers import get_product_block_names, sql_compile
from orchestrator.cli.domain_gen_helpers.product_block_helpers import get_product_block_id
from orchestrator.cli.domain_gen_helpers.types import DomainModelChanges
from orchestrator.cli.helpers.input_helpers import get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, print_fmt, str_fmt
from orchestrator.db import db
from orchestrator.db.models import (
    ProcessSubscriptionTable,
    ProcessTable,
    ProductTable,
    SubscriptionInstanceTable,
    SubscriptionTable,
    product_product_block_association,
)
from orchestrator.domain.base import SubscriptionModel, get_depends_on_product_block_type_list


def get_product_id(product_name: str) -> ScalarSelect:
    return get_product_ids([product_name])


def get_product_ids(product_names: list[str] | set[str]) -> ScalarSelect:
    return select(ProductTable.product_id).where(ProductTable.name.in_(product_names)).scalar_subquery()


def map_product_additional_relations(changes: DomainModelChanges) -> DomainModelChanges:
    """Map additional relations for created products.

    Adds resource type and product block relations.

    Args:
        changes: DomainModelChanges class with all changes.

    Returns: Updated DomainModelChanges.
    """

    for product_name, product_class in changes.create_products.items():
        for field_name in product_class._non_product_block_fields_.keys():
            changes.create_product_fixed_inputs.setdefault(field_name, set()).add(product_name)

        product_blocks_in_model = product_class._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)
        for product_block_name in get_product_block_names(product_blocks_types_in_model):
            changes.create_product_to_block_relations.setdefault(product_block_name, set()).add(product_name)
    return changes


def generate_create_products_sql(
    create_products: dict[str, type[SubscriptionModel]], inputs: dict[str, dict[str, str]]
) -> list[str]:
    """Generate SQL to create products.

    Args:
        create_products: Dict of SubscriptionModels by product name.
            - key: product name.
            - value: SubscriptionModel
        inputs: Optional Dict to add prefilled values for 'description', 'product_type' and 'tag' per product.
            - key: product name.
            - value: Dict with 'description', 'product_type' and 'tag'.
                - key: product property.
                - value: value for the property.

    Returns: List of SQL strings to create products.
    """
    print_fmt("\nCreate new products", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def create_product(product_name: str, product_type: str) -> str:
        values = inputs.get(product_name, {})
        print(f"Product: {product_type} {str_fmt(product_name, flags=[COLOR.BOLD])}")  # noqa: T001, T201
        description = values.get("description") or get_user_input("Supply the product description: ")
        tag = values.get("tag") or get_user_input("Supply the product tag: ")
        return sql_compile(
            Insert(ProductTable).values(
                {
                    "name": product_name,
                    "description": description,
                    "product_type": product_type,
                    "tag": tag,
                    "status": "active",
                }
            )
        )

    return [
        create_product(product_name, product_type.__name__) for product_name, product_type in create_products.items()
    ]


def generate_delete_products_sql(delete_products: set[str]) -> list[str]:
    """Generate SQL to delete products.

    Args:
        delete_products: List of product names.

    Returns: List of SQL strings to delete products.
    """
    if not delete_products:
        return []

    def delete_product_relations_sql(product_names: set[str]) -> list[str]:
        product_ids = get_product_ids(product_names)
        subscription_ids = (
            select(SubscriptionTable.subscription_id)
            .where(SubscriptionTable.product_id.in_(product_ids))
            .scalar_subquery()
        )
        process_ids = (
            select(ProcessSubscriptionTable.process_id)
            .where(ProcessSubscriptionTable.subscription_id.in_(subscription_ids))
            .scalar_subquery()
        )
        return [
            sql_compile(Delete(ProcessTable).where(ProcessTable.process_id.in_(process_ids))),
            sql_compile(
                Delete(ProcessSubscriptionTable).where(ProcessSubscriptionTable.subscription_id.in_(subscription_ids))
            ),
            sql_compile(
                Delete(SubscriptionInstanceTable).where(SubscriptionInstanceTable.subscription_id.in_(subscription_ids))
            ),
            sql_compile(Delete(SubscriptionTable).where(SubscriptionTable.product_id.in_(product_ids))),
        ]

    def delete_products_sql(product_names: set[str]) -> str:
        return sql_compile(Delete(ProductTable).where(ProductTable.name.in_(product_names)))

    sql_deletes = delete_product_relations_sql(delete_products)
    sql_deletes.append(delete_products_sql(delete_products))
    return sql_deletes


def generate_create_product_relations_sql(create_block_relations: dict[str, set[str]]) -> list[str]:
    """Generate SQL to create product relations to product blocks.

    Args:
        create_block_relations: Dict with product names by product block
            - key: product block name.
            - value: Set of product names to relate with.

    Returns: List of SQL strings to create subscription instances.
    """

    def create_block_relation(block_name: str, product_names: set[str]) -> str:
        block_id_sql = get_product_block_id(block_name)

        def create_block_relation_dict(product_name: str) -> dict[str, ScalarSelect]:
            product_id_sql = get_product_id(product_name)
            return {"product_block_id": block_id_sql, "product_id": product_id_sql}

        product_product_block_relation_dicts = [
            create_block_relation_dict(product_name) for product_name in product_names
        ]
        return sql_compile(Insert(product_product_block_association).values(product_product_block_relation_dicts))

    return [create_block_relation(*item) for item in create_block_relations.items()]


def generate_create_product_instance_relations_sql(product_to_block_relations: dict[str, set[str]]) -> list[str]:
    """Generate SQL to create subscription instances for existing subscriptions.

    Args:
        product_to_block_relations: Dict with product blocks by resource type
            - key: product block name.
            - value: Set of product names to relate with.

    Returns: List of SQL strings to create subscription instances.
    """

    def create_subscription_instance_relations(block_name: str, product_names: set[str]) -> Generator[str, None, None]:
        product_block_id = get_product_block_id(block_name)

        def map_subscription_instance_relations(
            product_name: str,
        ) -> Generator[dict[str, str | ScalarSelect], None, None]:
            product_id_sql = get_product_id(product_name)
            subscription_ids_stmt = select(SubscriptionTable.subscription_id).where(
                SubscriptionTable.product_id.in_(product_id_sql)
            )
            subscription_ids = db.session.scalars(subscription_ids_stmt)

            for subscription_id in subscription_ids:
                yield {"subscription_id": subscription_id, "product_block_id": product_block_id}

        subscription_relation_dicts = list(
            chain.from_iterable(map_subscription_instance_relations(block_name) for block_name in product_names)
        )
        if subscription_relation_dicts:
            yield sql_compile(Insert(SubscriptionInstanceTable).values(subscription_relation_dicts))

    return list(
        chain.from_iterable(
            create_subscription_instance_relations(*item) for item in product_to_block_relations.items()
        )
    )


def generate_delete_product_relations_sql(delete_block_relations: dict[str, set[str]]) -> list[str]:
    """Generate SQL to delete product to product blocks relations.

    Args:
        delete_block_relations: Dict with product blocks by resource type
            - key: product block name.
            - value: Set of product names to relate with.

    Returns: List of SQL strings to delete relations between product and product block.
    """

    def delete_block_relation(delete_block_name: str, product_names: set[str]) -> str:
        product_ids_sql = get_product_ids(product_names)
        block_id_sql = get_product_block_id(delete_block_name)
        return sql_compile(
            Delete(product_product_block_association).where(
                product_product_block_association.c.product_id.in_(product_ids_sql),
                product_product_block_association.c.product_block_id.in_(block_id_sql),
            )
        )

    return [delete_block_relation(*item) for item in delete_block_relations.items()]



from http import HTTPStatus
from uuid import uuid4

import pytest
from sqlalchemy import delete

from orchestrator.db import ProductBlockTable, db

PRODUCT_BLOCK_ID = uuid4()


@pytest.fixture
def seed():
    db.session.execute(delete(ProductBlockTable))
    product_block = ProductBlockTable(
        product_block_id=PRODUCT_BLOCK_ID,
        name="Product block 123",
        description="Original description of the product block",
        tag="tag123",
        status="active",
    )
    db.session.add(product_block)
    db.session.commit()


def test_product_block_by_id_404(seed, test_client):
    response = test_client.get(f"/api/product_blocks/{str(uuid4())}")
    assert HTTPStatus.NOT_FOUND == response.status_code


def test_update_description(seed, test_client):
    body = {
        "description": "BLABLA",
    }
    response = test_client.patch(f"/api/product_blocks/{PRODUCT_BLOCK_ID}", json=body)

    assert response.json()["description"] == "BLABLA"


def test_update_description_with_empty_body(seed, test_client):
    get_before = test_client.get(f"/api/product_blocks/{PRODUCT_BLOCK_ID}")
    body = {}
    response = test_client.patch(f"/api/product_blocks/{PRODUCT_BLOCK_ID}", json=body)

    assert HTTPStatus.CREATED == response.status_code
    get_after = test_client.get(f"/api/product_blocks/{PRODUCT_BLOCK_ID}")

    assert get_before.json()["description"] == get_after.json()["description"]


def test_update_description_nonexistent_product_block(seed, test_client):
    random_uuid = uuid4()
    body = {
        "description": "BLABLA",
    }
    response = test_client.patch(f"/api/product_blocks/{random_uuid}", json=body)

    assert HTTPStatus.NOT_FOUND == response.status_code



from sqlalchemy.inspection import inspect

from orchestrator.db import ProductTable
from orchestrator.db.filters import create_memoized_field_list
from orchestrator.db.sorting import generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

PRODUCT_SORT_FUNCTIONS_BY_COLUMN = {
    to_camel(key): generic_column_sort(value, ProductTable) for [key, value] in inspect(ProductTable).columns.items()
}

product_sort_fields = create_memoized_field_list(PRODUCT_SORT_FUNCTIONS_BY_COLUMN)
sort_products = generic_sort(PRODUCT_SORT_FUNCTIONS_BY_COLUMN)



from orchestrator.db import db
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.conftest import CUSTOMER_ID


def test_subscription_metrics_active_subscription(test_client, product_type_1_subscriptions_factory) -> None:
    product_type_1_subscriptions_factory()
    expected_metric = (
        f'wfo_subscriptions_count{{customer_id="{CUSTOMER_ID}",insync="True",lifecycle_state="active",'
        f'product_name="Product 1"}} 1.0'
    )
    response = test_client.get("/api/metrics")
    assert expected_metric in response.text

    product_type_1_subscriptions_factory(4)

    expected_metric = (
        f'wfo_subscriptions_count{{customer_id="{CUSTOMER_ID}",insync="True",lifecycle_state="active",'
        f'product_name="Product 1"}} 5.0'  # We now expect 5 subscriptions of the same type.
    )
    response = test_client.get("/api/metrics")
    assert expected_metric in response.text


def test_subscription_metrics_subscription_lifecycles(test_client, generic_subscription_1) -> None:
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    for lifecycle_state in SubscriptionLifecycle.values():
        subscription = SubscriptionModel.from_other_lifecycle(subscription, lifecycle_state)
        subscription.save()
        db.session.commit()
        expected_metric = (
            f'wfo_subscriptions_count{{customer_id="{CUSTOMER_ID}",insync="True",lifecycle_state="{lifecycle_state}",'
            f'product_name="Product 1"}} 1.0'
        )
        response = test_client.get("/api/metrics")
        assert expected_metric in response.text


def test_subscription_metrics_subscription_sync(test_client, generic_subscription_1) -> None:
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    subscription.insync = False
    subscription.save()
    db.session.commit()
    expected_metric = (
        f'wfo_subscriptions_count{{customer_id="{CUSTOMER_ID}",insync="False",lifecycle_state="active",'
        f'product_name="Product 1"}} 1.0'
    )
    response = test_client.get("/api/metrics")
    assert expected_metric in response.text

    subscription.insync = True
    subscription.save()
    db.session.commit()
    expected_metric = (
        f'wfo_subscriptions_count{{customer_id="{CUSTOMER_ID}",insync="True",lifecycle_state="active",'
        f'product_name="Product 1"}} 1.0'
    )
    response = test_client.get("/api/metrics")
    assert expected_metric in response.text



from typing import Optional

import pytest

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.helpers import safe_delete_product_block_id


# Note: don't use these classes directly. Use the `test_product_block_one_nested` fixture to ensure proper teardown.
class ProductBlockOneNestedForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneNestedForTest"):
    sub_block: Optional["ProductBlockOneNestedForTestInactive"] = None
    int_field: int | None = None


class ProductBlockOneNestedForTestProvisioning(
    ProductBlockOneNestedForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    sub_block: Optional["ProductBlockOneNestedForTestProvisioning"] = None
    int_field: int


class ProductBlockOneNestedForTest(ProductBlockOneNestedForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    sub_block: Optional["ProductBlockOneNestedForTest"] = None
    int_field: int


@pytest.fixture
def test_product_block_one_nested():
    # Classes defined at module level, otherwise they remain in local namespace and
    # `get_type_hints()` can't evaluate the ForwardRefs
    blocks = (
        ProductBlockOneNestedForTestInactive,
        ProductBlockOneNestedForTestProvisioning,
        ProductBlockOneNestedForTest,
    )

    yield blocks

    for block in blocks:
        safe_delete_product_block_id(block)


@pytest.fixture
def test_product_block_one_nested_db_in_use_by_block(resource_type_list, resource_type_int, resource_type_str):
    in_use_by_block = ProductBlockTable(
        name="ProductBlockOneNestedForTest", description="Test Block Parent", tag="TEST", status="active"
    )
    in_use_by_block.resource_types = [resource_type_int]

    db.session.add(in_use_by_block)
    db.session.commit()

    return in_use_by_block



from uuid import uuid4

from orchestrator.db import ProductTable, db
from orchestrator.types import SubscriptionLifecycle

from products.product_types.example2 import Example2, Example2Inactive


def test_example2_new():
    product = ProductTable.query.filter(ProductTable.name == "example2").one()

    diff = Example2.diff_product_in_database(product.product_id)
    assert diff == {}

    example2 = Example2Inactive.from_product_id(
        product_id=product.product_id,
        customer_id=uuid4(),
        status=SubscriptionLifecycle.INITIAL,
    )

    assert example2.subscription_id is not None
    assert example2.insync is False

    # TODO: Add more product specific asserts

    assert example2.description == f"Initial subscription of {product.description}"
    example2.save()

    example22 = Example2Inactive.from_subscription(example2.subscription_id)
    assert example2 == example22


def test_example2_load_and_save_db(example2_subscription):
    example2 = Example2.from_subscription(example2_subscription)

    assert example2.insync is True

    # TODO: Add more product specific asserts

    example2.description = "Changed description"

    # TODO: add a product specific change

    example2.save()

    # Explicit commit here as we are not running in the context of a step
    db.session.commit()

    example2 = Example2.from_subscription(example2_subscription)

    # TODO: Add more product specific asserts

    assert example2.description == "Changed description"
